from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QLabel,
    QCheckBox,
    QScrollArea,
)

from aydin.gui._qt.custom_widgets.constructor_arguments import (
    ConstructorArgumentsWidget,
)
from aydin.gui._qt.custom_widgets.vertical_line_break_widget import (
    QVerticalLineBreakWidget,
)
from aydin.restoration.denoise.util.denoise_utils import get_denoiser_class_instance


class DenoiseTabMethodWidget(QWidget):
    def __init__(self, parent, name=None, description=None):
        super(QWidget, self).__init__(parent)

        self.parent = parent
        self.name = name
        self.description = description

        # Widget layout
        self.layout = QHBoxLayout()
        self.tab_method_layout = QVBoxLayout()
        self.tab_method_layout.setAlignment(Qt.AlignTop)

        # Description Label
        self.description_scroll = QScrollArea()
        self.description_scroll.setStyleSheet("QScrollArea {border: none;}")
        self.description_scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.description_scroll.setAlignment(Qt.AlignTop)
        self.description_label = QLabel(self.description)
        self.description_label.setWordWrap(True)

        self.description_label.setTextFormat(Qt.RichText)
        self.description_label.setOpenExternalLinks(True)

        self.description_label.setAlignment(Qt.AlignTop)
        self.description_scroll.setWidget(self.description_label)
        self.description_scroll.setWidgetResizable(True)
        self.description_scroll.setMinimumHeight(300)

        self.tab_method_layout.addWidget(self.description_scroll)

        self.right_side_vlayout = QVBoxLayout()
        self.right_side_vlayout.setAlignment(Qt.AlignTop)

        # Checkboxes
        self.save_json_and_model_layout = QHBoxLayout()
        self.save_json_and_model_layout.setAlignment(Qt.AlignLeft)

        self.save_json_checkbox = QCheckBox("Save denoising options (JSON)")
        self.save_json_checkbox.setChecked(True)
        self.save_json_and_model_layout.addWidget(self.save_json_checkbox)
        self.save_json_and_model_layout.addWidget(QVerticalLineBreakWidget(self))

        self.save_model_checkbox = QCheckBox("Save the trained model")
        self.save_model_checkbox.setChecked(True)
        self.save_json_and_model_layout.addWidget(self.save_model_checkbox)

        self.right_side_vlayout.addLayout(self.save_json_and_model_layout)

        # Arguments
        self.scroll = QScrollArea()
        self.scroll.setStyleSheet("QScrollArea {border: none;}")
        self.scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.table_and_panes_widget = QWidget()
        self.table_and_panes_layout = QVBoxLayout()
        self.table_and_panes_widget.setLayout(self.table_and_panes_layout)
        self.scroll.setWidget(self.table_and_panes_widget)
        self.scroll.setWidgetResizable(True)
        self.scroll.setMinimumHeight(300)

        self.method_and_approach, self.implementation = self.name.split("-")

        args = get_denoiser_class_instance(variant=self.name).configurable_arguments[
            self.method_and_approach + "-" + self.implementation
        ]

        self.constructor_arguments_widget_dict = {}

        for component in list(args.keys()):
            sub_dict = args[component]

            self.constructor_arguments_widget = ConstructorArgumentsWidget(
                self,
                arg_names=sub_dict["arguments"],
                arg_defaults=sub_dict["defaults"],
                arg_annotations=sub_dict["annotations"],
                reference_class=sub_dict["reference_class"],
            )
            self.constructor_arguments_widget_dict[
                component
            ] = self.constructor_arguments_widget
            self.table_and_panes_layout.addWidget(self.constructor_arguments_widget)
            self.table_and_panes_layout.setSpacing(0)
            self.table_and_panes_layout.setAlignment(Qt.AlignTop)

        self.right_side_vlayout.addWidget(self.scroll)

        self.layout.addLayout(self.tab_method_layout, 35)
        self.layout.addWidget(QVerticalLineBreakWidget(self))
        self.layout.addLayout(self.right_side_vlayout, 50)

        self.setLayout(self.layout)

    def lower_level_args(self):
        args = {}

        for key, value in self.constructor_arguments_widget_dict.items():
            args[key] = value.params_dict

        args["variant"] = self.name

        return args



# flake8: noqa
import time

import numpy
import numpy as np
from skimage.data import astronaut
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import rgbtest
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo(image):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True

    amplitude = 190
    noisy = image.astype(np.int16) + np.random.randint(
        -amplitude, amplitude, size=image.shape, dtype=np.int16
    )
    noisy = noisy.clip(0, 255).astype(np.uint8)

    # image = image.astype(np.float32)
    # noisy = noisy.astype(np.float32)

    generator = StandardFeatureGenerator(include_spatial_features=True)
    regressor = LGBMRegressor(learning_rate=0.01)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(noisy, channel_axes=(False, False, True))
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    denoised = it.translate(noisy, channel_axes=(False, False, True))
    stop = time.time()
    print(f"Inference: elapsed time:  {stop - start} ")

    image = numpy.clip(image, 0, 255)
    noisy = numpy.clip(noisy, 0, 255)
    denoised = numpy.clip(denoised, 0, 255)

    print(
        "noisy",
        psnr(image, noisy, data_range=255),
        ssim(noisy, image, multichannel=True),
    )
    print(
        "denoised",
        psnr(image, denoised, data_range=255),
        ssim(denoised, image, multichannel=True),
    )
    # print("denoised_predict", psnr(denoised_predict, image), ssim(denoised_predict, image))

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image', rgb=True)
        viewer.add_image(noisy, name='noisy', rgb=True)
        viewer.add_image(denoised, name='denoised', rgb=True)

        cm = {0: 'r', 1: 'g', 2: 'b'}
        for i in range(3):
            viewer.add_image(image[..., i], name=f'image {cm[i]}')
            viewer.add_image(noisy[..., i], name=f'noisy {cm[i]}')
            viewer.add_image(denoised[..., i], name=f'denoised {cm[i]}')


image = rgbtest()
demo(image)
image = astronaut()
demo(image)



import numpy
from skimage.data import binary_blobs

from aydin.util.array.nd import (
    nd_range,
    nd_range_bool_tuple,
    nd_loop,
    nd_split_slices,
    remove_margin_slice,
    nd_range_radii,
    extract_tiles,
)


def test_nd_range():

    print(str(list(nd_range(-1, 1, 3))))
    assert (
        str(list(nd_range(-1, 1, 3)))
        == "[(-1, -1, -1), (-1, -1, 0), (-1, 0, -1), (-1, 0, 0), (0, -1, -1), (0, -1, 0), (0, 0, -1), (0, 0, 0)]"
    )


def test_nd_range_radii():

    radii = (1, 2)

    print(str(list(nd_range_radii(radii))))

    assert (
        str(list(nd_range_radii(radii)))
        == "[(-1, -2), (-1, -1), (-1, 0), (-1, 1), (-1, 2), (0, -2), (0, -1), (0, 0), (0, 1), (0, 2), (1, -2), (1, -1), (1, 0), (1, 1), (1, 2)]"
    )


def test_nd_range_bool_tuple():

    print(str(list(nd_range_bool_tuple(-1, 1, (True, True, False)))))
    assert (
        str(list(nd_range_bool_tuple(-1, 1, (True, True, False))))
        == "[(-1, -1, 0), (-1, 0, 0), (0, -1, 0), (0, 0, 0)]"
    )

    print(str(list(nd_range_bool_tuple(-1, 1, (False, True, False)))))
    assert (
        str(list(nd_range_bool_tuple(-1, 1, (False, True, False))))
        == "[(0, -1, 0), (0, 0, 0)]"
    )


def test_nd_loop():

    loop_tuple = (2, 1, 3, 5)

    print(str(list(nd_loop(loop_tuple))))
    assert (
        str(list(nd_loop(loop_tuple)))
        == "[(0, 0, 0, 0), (0, 0, 0, 1), (0, 0, 0, 2), (0, 0, 0, 3), (0, 0, 0, 4), (0, 0, 1, 0), (0, 0, 1, 1), (0, 0, 1, 2), (0, 0, 1, 3), (0, 0, 1, 4), (0, 0, 2, 0), (0, 0, 2, 1), (0, 0, 2, 2), (0, 0, 2, 3), (0, 0, 2, 4), (1, 0, 0, 0), (1, 0, 0, 1), (1, 0, 0, 2), (1, 0, 0, 3), (1, 0, 0, 4), (1, 0, 1, 0), (1, 0, 1, 1), (1, 0, 1, 2), (1, 0, 1, 3), (1, 0, 1, 4), (1, 0, 2, 0), (1, 0, 2, 1), (1, 0, 2, 2), (1, 0, 2, 3), (1, 0, 2, 4)]"
    )


def test_nd_split_slices():

    array_shape = (96, 17, 117, 45)
    nb_slices = (5, 3, 7, 5)

    slice_tuple_list = list(nd_split_slices(array_shape, nb_slices, do_shuffle=True))

    print(str(slice_tuple_list))

    array_source = numpy.random.choice(a=[False, True], size=array_shape)
    array_target = numpy.random.choice(a=[False, True], size=array_shape)

    for slice_tuple in slice_tuple_list:

        array_target[slice_tuple] = array_source[slice_tuple]
        pass

    assert numpy.all(array_source == array_target)


def test_nd_split_slices_with_margins():

    array_shape = (96, 17, 117, 45)
    nb_slices = (5, 3, 7, 5)
    margins = (4, 5, 6, 7)

    slice_tuple_list = list(nd_split_slices(array_shape, nb_slices, do_shuffle=False))
    slice_margin_tuple_list = list(
        nd_split_slices(array_shape, nb_slices, do_shuffle=False, margins=margins)
    )

    # print(str(slice_tuple_list))

    array_source = numpy.random.choice(a=[0, 1], size=array_shape)
    array_target = numpy.zeros(array_shape, dtype=numpy.int64)

    for slice_tuple, slice_margin_tuple in zip(
        slice_tuple_list, slice_margin_tuple_list
    ):

        sliced_array_with_margin = array_source[slice_margin_tuple]
        remove_margin_slice_tuple = remove_margin_slice(
            array_shape, slice_margin_tuple, slice_tuple
        )
        sliced_array_removed_margin = sliced_array_with_margin[
            remove_margin_slice_tuple
        ]
        # print(array_target[slice_tuple].shape)
        # print(sliced_array_removed_margin.shape)
        array_target[slice_tuple] += sliced_array_removed_margin

    assert numpy.all(array_source == array_target)


def test_extract_tiles():

    blob_image = binary_blobs(length=512, seed=1, n_dim=2)

    tiles = extract_tiles(blob_image, tile_size=64, extraction_step=64, flatten=True)

    print(tiles.shape)
    print(len(tiles))



import time

import numpy
import pytest
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim
from skimage.util import random_noise

from aydin.io.datasets import newyork
from aydin.regression.lgbm import LGBMRegressor


def normalise(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


class Timer:
    def __enter__(self):
        self.start = time.clock()
        return self

    def __exit__(self, *args):
        self.end = time.clock()
        self.interval = self.end - self.start


@pytest.mark.opencl
def test_opencl_inference():
    from aydin.features.fast.fast_features import FastFeatureGenerator

    image = newyork().astype(numpy.float32)
    image = normalise(image)

    intensity = 5
    numpy.random.seed(0)
    noisy = numpy.random.poisson(image * intensity) / intensity
    noisy = random_noise(noisy, mode='gaussian', var=0.01, seed=0)
    noisy = noisy.astype(numpy.float32)

    # feature generator requires images in 'standard' form: BCTZYX
    noisy = image[numpy.newaxis, numpy.newaxis, ...]

    generator = FastFeatureGenerator()
    regressor = LGBMRegressor(max_num_estimators=130)

    def callback(iteration, val_loss, model):
        print(f"iteration={iteration}, val_loss={val_loss}")

    print('feature generation...')
    features = generator.compute(noisy, exclude_center_value=True)

    print('reshaping...')
    x = features.reshape(-1, features.shape[-1])
    y = noisy.reshape(-1)

    print(f"Number of data points             : {x.shape[0]}")
    print(f"Number of features per data points: {x.shape[-1]}")

    print('fitting...')
    regressor.fit(x, y, x_valid=x, y_valid=y, regressor_callback=callback)

    print('feature generation for inference -- with center value --...')
    features = generator.compute(noisy, exclude_center_value=False)
    x = features.reshape(-1, features.shape[-1])

    for i in range(1, 20):
        print(f'predicting gpu ({i})...')
        regressor.gpu_prediction = True
        with Timer() as t:
            yp_gpu = regressor.predict(x)
        print(f' ... took {t.interval} seconds')

    # uncomment to compare speed (typically gpu = 20x faster than cpu)
    print('predicting cpu ... ')
    regressor.gpu_prediction = False
    with Timer() as t:
        yp_cpu = regressor.predict(x)
    print(f' ... took {t.interval} seconds')
    # yp_cpu = yp_gpu*0.01 # place holder array so that rest of the code runs

    denoised_cpu = yp_cpu.reshape(image.shape)
    denoised_cpu = numpy.clip(denoised_cpu, 0, 1)
    denoised_gpu = yp_gpu.reshape(image.shape)
    denoised_gpu = numpy.clip(denoised_gpu, 0, 1)

    ssim_value_cpu = ssim(denoised_cpu, image)
    psnr_value_cpu = psnr(denoised_cpu, image)
    ssim_value_gpu = ssim(denoised_gpu, image)
    psnr_value_gpu = psnr(denoised_gpu, image)

    print("denoised cpu", psnr_value_cpu, ssim_value_cpu)
    print("denoised gpu", psnr_value_gpu, ssim_value_gpu)

    assert ssim_value_gpu > 0.50



import numpy
from scipy.ndimage import gaussian_filter

from aydin.analysis.resolution_estimate import resolution_estimate
from aydin.io.datasets import normalise, add_noise, cropped_newyork
from aydin.util.log.log import Log


def demo_resolution_estimate(display: bool = True):
    Log.enable_output = True

    image = cropped_newyork()[:-3, 0:-7]
    high_res_image = normalise(image.astype(numpy.float32))
    low_res_image = gaussian_filter(high_res_image, sigma=3)
    high_res_noisy = add_noise(high_res_image)
    low_res_noisy = add_noise(low_res_image)

    fc_high_res, _ = resolution_estimate(high_res_image, display_images=display)
    fc_low_res, _ = resolution_estimate(low_res_image, display_images=display)
    fc_high_res_noisy, _ = resolution_estimate(high_res_noisy, display_images=display)
    fc_low_res_noisy, _ = resolution_estimate(low_res_noisy, display_images=display)

    print(f"fc_high_res={fc_high_res}")
    print(f"fc_low_res={fc_low_res}")
    print(f"fc_high_res_noisy={fc_high_res_noisy}")
    print(f"fc_low_res_noisy={fc_low_res_noisy}")

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='clean_image')
            viewer.add_image(high_res_image, name='high_res_image')
            viewer.add_image(low_res_image, name='low_res_image')
            viewer.add_image(high_res_noisy, name='high_res_noisy')
            viewer.add_image(low_res_noisy, name='low_res_noisy')

    assert fc_low_res < 0.6 * fc_high_res
    assert fc_low_res_noisy < 0.6 * fc_high_res_noisy


if __name__ == "__main__":
    demo_resolution_estimate()



import time

import numpy
import pytest
from skimage.data import binary_blobs
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import add_noise
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor
from aydin.regression.lgbm import LGBMRegressor
from aydin.regression.nn import NNRegressor
from aydin.util.log.log import Log

display_for_debug = False


@pytest.mark.heavy
def test_it_fgr_nn_2D():
    it_fgr_nD(2, 512, numpy.s_[..., 0:281, 0:413], regressor='nn', min_ssim=0.55)


@pytest.mark.heavy
def test_it_fgr_nn_3D():
    it_fgr_nD(3, 128, numpy.s_[..., 0:111, 0:113, 0:97], regressor='nn', min_ssim=0.70)


@pytest.mark.heavy
def test_it_fgr_nn_4D():
    it_fgr_nD(
        4, 64, numpy.s_[..., 0:11, 0:41, 0:57, 0:53], regressor='nn', min_ssim=0.70
    )


@pytest.mark.heavy
def test_it_fgr_gbm_2D():
    it_fgr_nD(2, 512, numpy.s_[..., 0:201, 0:213], regressor='gbm', min_ssim=0.55)


@pytest.mark.heavy
def test_it_fgr_gbm_3D():
    it_fgr_nD(3, 48, numpy.s_[..., 0:41, 0:43, 0:37], regressor='gbm', min_ssim=0.70)


@pytest.mark.heavy
def test_it_fgr_gbm_4D():
    it_fgr_nD(
        4, 48, numpy.s_[..., 0:11, 0:23, 0:22, 0:21], regressor='gbm', min_ssim=0.70
    )


@pytest.mark.heavy
def test_it_fgr_cb_2D():
    it_fgr_nD(2, 512, numpy.s_[..., 0:201, 0:213], regressor='cb', min_ssim=0.50)


@pytest.mark.heavy
def test_it_fgr_cb_3D():
    it_fgr_nD(3, 48, numpy.s_[..., 0:41, 0:43, 0:37], regressor='cb', min_ssim=0.70)


@pytest.mark.heavy
def test_it_fgr_cb_4D():
    it_fgr_nD(
        4, 48, numpy.s_[..., 0:11, 0:23, 0:22, 0:21], regressor='cb', min_ssim=0.70
    )


@pytest.mark.heavy
def test_it_fgr_gbm_2D_batchdims():
    Log.set_log_max_depth(2)
    it_fgr_nD(
        2,
        256,
        numpy.s_[..., 0:117, 0:175],
        regressor='gbm',
        batch_dims=(False, True),
        min_ssim=0.30,
    )


@pytest.mark.heavy
def test_it_fgr_gbm_3D_batchdims():
    Log.set_log_max_depth(2)
    it_fgr_nD(
        3,
        48,
        numpy.s_[..., 0:31, 0:37, 0:41],
        regressor='gbm',
        batch_dims=(False, True, False),
        min_ssim=0.80,
    )


@pytest.mark.heavy
def test_it_fgr_gbm_4D_batchdims():
    Log.set_log_max_depth(2)
    it_fgr_nD(
        4,
        48,
        numpy.s_[..., 0:11, 0:13, 0:17, 0:15],
        regressor='gbm',
        batch_dims=(False, True, False, True),
        min_ssim=0.66,
    )


def it_fgr_nD(
    n_dim,
    length=128,
    train_slice=numpy.s_[...],
    regressor='nn',
    batch_dims=None,
    min_ssim=0.85,
):
    """
    Test for self-supervised denoising using camera image with synthetic noise
    """

    image = binary_blobs(length=length, seed=1, n_dim=n_dim).astype(numpy.float32)
    image = n(image)

    noisy = add_noise(image)
    train = noisy[train_slice]

    generator = StandardFeatureGenerator()

    if regressor == 'nn':
        regressor = NNRegressor(max_epochs=10)
    elif regressor == 'gbm':
        regressor = LGBMRegressor()
    elif regressor == 'cb':
        regressor = CBRegressor()

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(train, train, batch_axes=batch_dims)
    stop = time.time()
    print(f"####### Training: elapsed time:  {stop - start} sec")

    start = time.time()
    denoised = it.translate(noisy, batch_axes=batch_dims)
    stop = time.time()
    print(f"####### Inference: elapsed time:  {stop - start} sec")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    # if the line below fails, then the parameters of the image the lgbm regressor have been broken.
    # do not change the number below, but instead, fix the problem -- most likely a parameter.

    assert ssim_denoised > min_ssim

    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )



from typing import Optional

import numpy

# from numpy.typing import ArrayLike

from aydin.it.normalisers.base import NormaliserBase
from aydin.it.normalisers.minmax import MinMaxNormaliser
from aydin.it.normalisers.percentile import PercentileNormaliser
from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class RangeTransform(ImageTransformBase):
    """Range Normalisation

    Images come in all sorts of formats, and pixels can be represented as 8,
    16, 32 bit integers as well as 16, 32, 64 bit floats. More crucially,
    the actual range of values can vary potentially causing difficulties for
    denoising algorithms. For these reasons and more, it is almost always
    recommended to normalise images to a range within [0, 1] and represented
    as 32 bit floats (sufficient for most cases and ideal for aydin).
    Finally, there are two different normalisation modes. The first 'minmax'
    simply finds the min and max values in the image and uses that to rescale
    to [0, 1]. However, there are sometimes outlier values that are isolated
    and completely out of context which would skew the normalisation.
    Therefore, we also have a 'percentile' mode that uses percentiles to
    determine the range -- typically 1% for min value and 99% for max value.
    Optionally, the image can be left as float after denormalisation,
    and values can be clipped to the range during both normalisation and
    denormalisation.
    """

    preprocess_description = (
        "Range normalisation" + ImageTransformBase.preprocess_description
    )
    postprocess_description = (
        "Range denormalisation" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = True

    def __init__(
        self,
        mode: str = 'minmax',
        percentile: Optional[float] = None,
        force_float_datatype: bool = False,
        clip: bool = True,
        priority: float = 0.2,
        **kwargs,
    ):

        """
        Constructs a Range Transform

        Parameters
        ----------
        mode : str
            Range normalisation mode: 'minmax' or 'percentile'
        percentile : Optional[float]
            Percentile value for the 'percentile' mode.
            If None the percentile value is automatically chosen
            based on the number of voxels in the image.
        force_float_datatype: bool
            After denormalisation the values are left as 32 bit
            floats instead of being converted back to the original
            data type. If False the best setting is automatically chosen.
        clip: bool
            Clips values outside of the range during normalisation
            and denormalisation.
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)

        self.mode = mode
        self.percentile = percentile
        self.force_float_datatype = force_float_datatype
        self.clip = clip
        self._normaliser: NormaliserBase
        self._min_value = None
        self._max_value = None

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_normaliser']
        del state['_min_value']
        del state['_max_value']
        return state

    def __str__(self):
        return (
            f'{type(self).__name__}'
            f' (mode={self.mode},'
            f' percentile={self.percentile},'
            f' leave_as_float={self.force_float_datatype},'
            f' clip={self.clip} )'
        )

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):

        with lsection(
            f"Normalizing value range ({self.mode}) for array of shape: {array.shape} and dtype: {array.dtype}"
        ):

            self._original_dtype = array.dtype
            array = array.astype(numpy.float32, copy=False)

            if self.mode == 'minmax':
                normaliser = MinMaxNormaliser()
            elif self.mode == 'percentile':
                normaliser = PercentileNormaliser(percentile=self.percentile)

            self._min_value, self._max_value = normaliser.calibrate(array)

            new_array = normaliser.normalise(array)

            self._normaliser = normaliser
            return new_array

    def postprocess(self, array):

        if not self.do_postprocess:
            return array

        with lsection(
            f"Denormalizing value range ({self.mode}) for array of shape: {array.shape} and dtype: {array.dtype}"
        ):
            force_float_datatype = self.force_float_datatype

            # Let's figure out if it is reasonable to keep the denoised data as float:
            if force_float_datatype is False and numpy.issubdtype(
                self._original_dtype, numpy.integer
            ):
                range = abs(self._max_value - self._min_value)
                if range < 128:
                    force_float_datatype = True

            new_array = self._normaliser.denormalise(
                array, leave_as_float=force_float_datatype, clip=self.clip
            )
            new_array = new_array.astype(self._original_dtype, copy=False)

            return new_array



from aydin.io.datasets import cropped_newyork

from aydin.it.classic_denoisers.demo.demo_2D_tv import demo_tv
from aydin.it.classic_denoisers.test.util_test_nd import check_nd
from aydin.it.classic_denoisers.tv import denoise_tv


def test_tv():
    assert demo_tv(cropped_newyork(), display=False) >= 0.627 - 0.03


def test_tv_nd():
    check_nd(denoise_tv)



from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import examples_single
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.attenuation import AttenuationTransform
from aydin.it.transforms.range import RangeTransform
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo_attenuation_correction_real():

    Log.override_test_exclusion = True
    Log.enable_output = True

    image = examples_single.royerlab_hcr.get_array().squeeze()
    image = image[2, :, :, :]

    ac = AttenuationTransform(axes=0)

    preprocessed = ac.preprocess(image)
    postprocessed = ac.postprocess(preprocessed)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(preprocessed, name='preprocessed')
        viewer.add_image(postprocessed, name='postprocessed')

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )
    regressor = LGBMRegressor(patience=128, compute_training_loss=True)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.add_transform(RangeTransform())
    it.train(image)
    denoised_without_preprocessing = it.translate(image)

    it.add_transform.append(ac)
    it.train(image)
    denoised_with_preprocessing = it.translate(image)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(preprocessed, name='preprocessed')
        viewer.add_image(postprocessed, name='postprocessed')
        viewer.add_image(
            denoised_without_preprocessing, name='denoised_without_preprocessing'
        )
        viewer.add_image(
            denoised_with_preprocessing, name='denoised_with_preprocessing'
        )


demo_attenuation_correction_real()



from copy import deepcopy
from os import listdir
from os.path import join, isfile, isdir
from pathlib import Path

from aydin.io import imread
from aydin.io.utils import split_image_channels, hyperstack_arrays
from aydin.util.log.log import lprint


class DataModel:
    """DataModel - manages our business logic regarding
    the images loaded into our GUI.

    Parameters
    ----------
    parent : Object
    """

    def __init__(self, parent=None):
        self.parent = parent
        self._filepaths = dict()
        self._images = []

    @property
    def filepaths(self):
        """filepaths is the dict which has file paths as keys
        and corresponding tuple of array and metadata as value.

        Returns
        -------
        self._filepaths : dict

        """
        return self._filepaths

    def clear_filepaths(self):
        """Clears the self._filepaths attribute and images of
        the model and triggers needed updates on files view.
        """
        self._filepaths = dict()
        self.update_files_tabview()
        self.clear_images()

    def add_filepaths(self, file_paths):
        """Adds filepaths to the model if they do not already
        exist in the current state of the model. If any file
        path is added, method also triggers addition of the
        corresponding images to the model too.

        Parameters
        ----------
        file_paths : List[str]

        """
        new_paths_added = False
        new_images = {}

        # If we drag and drop folders we add the files in the folders instead:
        if all(isdir(folder_path) for folder_path in file_paths):
            expanded_paths = []
            for folder_path in file_paths:
                for file_path in listdir(folder_path):
                    expanded_paths.append(join(folder_path, file_path))
            file_paths = expanded_paths

        # Add each path checking whether it is already present and an actual file:
        for file_path in file_paths:
            if file_path not in self._filepaths and isfile(file_path):
                array, metadata = imread(file_path)
                if array is None and metadata is None:
                    continue
                else:
                    new_paths_added = True
                    self._filepaths[file_path] = (array, metadata)
                    new_images[file_path] = (array, metadata)

        if new_paths_added:
            self.update_files_tabview()
            self.add_images(new_images)

    def remove_filepaths(self, fpaths):
        """Removes given file paths and their corresponding
        images from the model.

        Parameters
        ----------
        fpaths : List[str]

        """
        for fpath in fpaths:
            del self._filepaths[fpath]
            self.remove_image(fpath)

    @property
    def images(self):
        """images is the list of a list with a particular
        format for each existing image in the model.

        Returns
        -------
        self._images : List[List[str, numpy.typing.ArrayLike, FileMetadata, bool, bool, str]]

        """
        return self._images

    def clear_images(self):
        """Clears the self._images attribute and triggers
        needed updates on images view.
        """
        self._images.clear()
        self.update_images_tabview()

    def add_images(self, new_images, train_on=True, denoise=True):
        """Adds images to the model. Method also triggers
        needed updates on images view.

        Parameters
        ----------
        new_images : dict
            Expects a particular dict format where key is
            the filepath and value is tuple of corresponding
            array and metadata.
        train_on : bool, optional
        denoise : bool, optional

        """
        for path, (array, metadata) in new_images.items():
            self._images.append(
                [Path(path).name, array, metadata, train_on, denoise, path]
            )

        self.update_images_tabview()

    def remove_image(self, image_filepath):
        """Removes all images from the model for the given
        file path.

        Parameters
        ----------
        image_filepath : str

        """
        indices2remove = []
        for idx, imagelist_item in enumerate(self._images):
            if (
                Path(image_filepath).parents[0] == Path(imagelist_item[5]).parents[0]
                and Path(image_filepath).name in imagelist_item[0]
            ):
                indices2remove.append(idx)

        for idx in reversed(indices2remove):
            self._images.pop(idx)

        self.update_images_tabview()

    @property
    def images_to_denoise(self):
        """Returns sublist of self.images with only elements
        of it that are marked to be denoised.

        Returns
        -------
        List[List[str, numpy.typing.ArrayLike, FileMetadata, bool, bool, str]]

        """
        return list(filter(lambda image: image[4], self.images))

    def set_image_to_denoise(self, filename, new_value):
        """Method to mark/unmark to denoise an image the with
        corresponding filename.

        Parameters
        ----------
        filename : str
        new_value : bool

        """
        for imagelist_item in self._images:
            if imagelist_item[0] == filename:
                imagelist_item[4] = new_value
                self.update_dimensions_tabview()
                self.update_cropping_tabview()

    def set_split_channels(self, filename, filepath, new_value: bool):
        """Method to split/de-split channels of an image with
        corresponding filename.

        Parameters
        ----------
        filename : str
        filepath : str
        new_value : bool

        Returns
        -------
        int
            returns -1 if the image doesn't have a channel axis.

        """
        # Fetch the self._images elements that are associated by their file names
        imagelist_items = [elem for elem in self._images if filename in elem[0]]

        if new_value:  # if the split checkbox value is True, we try to split

            # There should be only one related image as a non-splitted file
            if len(imagelist_items) != 1:
                return

            imagelist_item = imagelist_items[0]

            if imagelist_item[2].axes.find("C") == -1:
                lprint("Array has no channel axis detected")
                return -1

            self.remove_image(filepath)  # Remove the existing image from the model

            # Split the array and collect new arrays and metadatas
            splitted_arrays, metadatas = split_image_channels(
                imagelist_item[1], deepcopy(imagelist_item[2])
            )

            # Prepare new file names for the splitted images
            filenames = [f"channel_{_}_{filename}" for _ in range(len(splitted_arrays))]

            # Add the splitted images to the model
            new_images = {}
            for filename, array, metadata in zip(filenames, splitted_arrays, metadatas):
                new_images[filepath.replace(Path(filepath).name, filename)] = (
                    array,
                    metadata,
                )
            self.add_images(new_images)

            self.update_images_tabview()  # Update the images view
        else:  # if the split checkbox value is False, we try to de-split
            if (
                len(imagelist_items) > 1
            ):  # There should be multiple related splitted images
                self.remove_image(filepath)  # Remove the existing splitted images

                self.add_images(
                    {filepath: self.filepaths[filepath]}
                )  # Add back to original image

                self.update_images_tabview()  # update the images view

    def set_hyperstack(self, hyperstack_checkbox_value: bool):
        """Method to hyperstack/de-hyperstack files in the model.

        Parameters
        ----------
        hyperstack_checkbox_value : bool

        Returns
        -------
        int
            returns -1 if hyperstacking fails for any reason.

        """
        if hyperstack_checkbox_value:  # We try to hyperstack
            try:
                # Gather arrays and metadatas from all files
                arrays = []
                metadatas = []

                for path, (array, metadata) in self.filepaths.items():
                    arrays.append(array)
                    metadatas.append(metadata)

                # Hyperstack those
                hyperstacked_array, hyperstacked_metadata = hyperstack_arrays(
                    arrays, metadatas
                )

                # Prepare file name and path for the hyperstacked image
                hyperstacked_filename = f"hyperstack_{self.images[0][0]}"
                filepath = self.images[0][5]
                hyperstacked_filepath = filepath.replace(
                    Path(filepath).name, hyperstacked_filename
                )

                # Update images
                self.clear_images()  # Remove old images
                self.add_images(
                    {hyperstacked_filepath: (hyperstacked_array, hyperstacked_metadata)}
                )  # Add the hyperstacked image

                lprint("Images stacked")
            except Exception as e:
                print(e)
                return -1
        else:  # We try to de-hyperstack
            filepaths = self.filepaths  # Keep the original file list
            self.clear_filepaths()  # Clear all the files and images
            self.add_filepaths(filepaths)  # Add the original file list back

    def update_files_tabview(self):
        """Method to trigger updating files tab view."""
        self.parent.filestab_changed()

    def update_images_tabview(self):
        """Method to trigger updating images tab view."""
        self.parent.imagestab_changed()
        self.update_dimensions_tabview()
        self.update_cropping_tabview()

    def update_dimensions_tabview(self):
        """Method to trigger updating dimensions tab view."""
        self.parent.dimensionstab_changed()

    def update_cropping_tabview(self):
        """Method to trigger updating cropping tab view."""
        self.parent.croppingtabs_changed()



# flake8: noqa
import os
import time

import numpy
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, add_noise, pollen, newyork, lizard, characters
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.rf import RFRegressor
from aydin.util.log.log import Log


def demo(image, name):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """

    Log.enable_output = True

    image = normalise(image.astype(numpy.float32))
    noisy = add_noise(image)

    generator = StandardFeatureGenerator()

    regressor = RFRegressor(patience=32)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    denoised = it.translate(noisy)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)
    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)
    print("noisy   :", psnr_noisy, ssim_noisy)
    print("denoised:", psnr_denoised, ssim_denoised)

    # import matplotlib.pyplot as plt
    #
    # plt.figure(figsize=(2.7 * 5, 5), dpi=300)
    # plt.subplot(1, 3, 1)
    # plt.imshow(normalise(noisy), cmap='gray')
    # plt.axis('off')
    # plt.title(f'Noisy \nPSNR: {psnr_noisy:.3f}, SSIM: {ssim_noisy:.3f}')
    # plt.subplot(1, 3, 2)
    # plt.imshow(normalise(denoised), cmap='gray')
    # plt.axis('off')
    # plt.title(f'Denoised \nPSNR: {psnr_denoised:.3f}, SSIM: {ssim_denoised:.3f}')
    # plt.subplot(1, 3, 3)
    # plt.imshow(normalise(image), cmap='gray')
    # plt.axis('off')
    # plt.title('Original')
    # plt.subplots_adjust(left=0.01, right=0.99, top=0.95, bottom=0.01, hspace=0.1)
    # os.makedirs("../demo_results", exist_ok=True)
    # plt.savefig(f'demo_results/n2s_rf_2D_{name}.png')

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(normalise(image), name='image')
        viewer.add_image(normalise(noisy), name='noisy')
        viewer.add_image(normalise(denoised), name='denoised')


newyork_image = newyork()
demo(newyork_image, "newyork")
characters_image = characters()
demo(characters_image, "characters")

camera_image = camera()
demo(camera_image, "camera")
lizard_image = lizard()
demo(lizard_image, "lizard")
pollen_image = pollen()
demo(pollen_image, "pollen")


fibsem_image = fibsem()
demo(fibsem_image, "fibsem")



from mock import Mock

from aydin.gui.tabs.data_model import DataModel
from aydin.io import imread
from aydin.io.datasets import examples_single
from aydin.io.io import FileMetadata


def test_not_adding_non_existing_filepaths():
    data_model = DataModel(Mock())
    data_model.add_filepaths(["/asdfas/asdfasd/fdsaf/fas", "/fdsa/fdsa/fdsa/fdsa"])

    assert data_model.filepaths == {}


def test_adding_files():
    data_model = DataModel(Mock())
    fpath = examples_single.fountain.get_path()
    data_model.add_filepaths([fpath])
    array, metadata = imread(fpath)
    target_dict = {fpath: (array, metadata)}

    for key in data_model.filepaths.keys():
        for elem1, elem2 in zip(data_model.filepaths[key], target_dict[key]):
            if isinstance(elem1, FileMetadata):
                assert elem1 == elem2
            else:
                assert elem1.all() == elem2.all()


def test_removing_files():
    data_model = DataModel(Mock())
    fpath = examples_single.fountain.get_path()
    data_model.add_filepaths([fpath])

    data_model.clear_filepaths()

    assert data_model.filepaths == dict()


def test_removing_multiple_but_not_all_files():
    pass


def test_hyperstack():
    pass


def test_hyperstack_attempt_images_of_different_shapes():
    pass



import torch
import torch.nn as nn
import torch.nn.functional as F


class BasicConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(BasicConvBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1)
        self.norm = nn.InstanceNorm2d(out_channels)

    def forward(self, x):
        x = F.leaky_relu(
            self.norm(self.conv2(self.conv1(x))), negative_slope=0.001, inplace=True
        )
        return x


class BasicUNet(nn.Module):
    def __init__(self, n_input_channel=1, n_output_channel=1, nic=8, residual=True):
        super().__init__()

        self.residual = residual

        self.down1 = nn.MaxPool2d(kernel_size=2)
        self.down2 = nn.MaxPool2d(kernel_size=2)
        self.down3 = nn.MaxPool2d(kernel_size=2)
        self.down4 = nn.MaxPool2d(kernel_size=2)

        self.up1 = lambda x: nn.functional.interpolate(
            x, mode='nearest', scale_factor=2
        )
        self.up2 = lambda x: nn.functional.interpolate(
            x, mode='nearest', scale_factor=2
        )
        self.up3 = lambda x: nn.functional.interpolate(
            x, mode='nearest', scale_factor=2
        )
        self.up4 = lambda x: nn.functional.interpolate(
            x, mode='nearest', scale_factor=2
        )

        self.conv1 = BasicConvBlock(n_input_channel, nic)

        self.conv2 = BasicConvBlock(nic, nic * 2)
        self.conv3 = BasicConvBlock(nic * 2, nic * 4)
        self.conv4 = BasicConvBlock(nic * 4, nic * 8)
        self.conv5 = BasicConvBlock(nic * 8, nic * 16)

        self.conv6 = BasicConvBlock(nic * 3 * 8, nic * 8)
        self.conv7 = BasicConvBlock(nic * 3 * 4, nic * 4)
        self.conv8 = BasicConvBlock(nic * 3 * 2, nic * 2)
        self.conv9 = BasicConvBlock(nic * 3, nic)

        self.conv10 = BasicConvBlock(nic, n_output_channel)

        if self.residual:
            self.conv11 = BasicConvBlock(n_input_channel, n_output_channel)

    def forward(self, x):
        c0 = x
        c1 = self.conv1(x)
        x = self.down1(c1)
        c2 = self.conv2(x)
        x = self.down2(c2)
        c3 = self.conv3(x)
        x = self.down3(c3)
        c4 = self.conv4(x)
        x = self.down4(c4)
        x = self.conv5(x)
        x = self.up1(x)
        x = torch.cat([x, c4], 1)
        x = self.conv6(x)
        x = self.up2(x)
        x = torch.cat([x, c3], 1)
        x = self.conv7(x)
        x = self.up3(x)
        x = torch.cat([x, c2], 1)
        x = self.conv8(x)
        x = self.up4(x)
        x = torch.cat([x, c1], 1)
        x = self.conv9(x)
        x = self.conv10(x)
        if self.residual:
            x = torch.add(x, self.conv11(c0))

        return x



# flake8: noqa
from aydin.util.optimizer.demo.demo_optimizer import (
    demo_optimizer_beale,
    demo_optimizer_mccormik,
    demo_optimizer_booth,
    demo_optimizer_sphere,
    demo_optimizer_rosenbrock,
    demo_optimizer_goldsteinprice,
    demo_optimizer_matyas,
    demo_optimizer_himmelblau,
)


def test_optimizer_beale():
    demo_optimizer_beale()


def test_optimizer_mccormik():
    demo_optimizer_mccormik()


def test_optimizer_booth():
    demo_optimizer_booth()


def test_optimizer_sphere():
    demo_optimizer_sphere()


def test_optimizer_rosenbrock():
    demo_optimizer_rosenbrock()


def test_optimizer_goldsteinprice():
    demo_optimizer_goldsteinprice()


def test_optimizer_matyas():
    demo_optimizer_matyas()


def test_optimizer_himmelblau():
    demo_optimizer_himmelblau()



from pprint import pprint

from aydin.restoration.denoise.util.denoise_utils import (
    get_list_of_denoiser_implementations,
)


def est_denoise_utils():
    implementations = get_list_of_denoiser_implementations()
    pprint(implementations)



from aydin.io.datasets import cropped_newyork

from aydin.it.classic_denoisers.demo.demo_2D_gaussian import demo_gaussian
from aydin.it.classic_denoisers.gaussian import denoise_gaussian
from aydin.it.classic_denoisers.test.util_test_nd import check_nd


def test_gaussian():
    assert demo_gaussian(cropped_newyork(), display=False) >= 0.600 - 0.02


def test_gaussian_nd():
    check_nd(denoise_gaussian)



from aydin.util.fast_uniform_filter.numba_cpu_uf import numba_cpu_uniform_filter
from aydin.util.fast_uniform_filter.parallel_uf import parallel_uniform_filter
from aydin.util.log.log import lprint


def uniform_filter_auto(image, size: int, printout_choice: bool = False):

    # Ensures size parameter is integer:
    size = int(round(size))

    # Different methods perform differently based on filter size:
    max_size = max(size) if isinstance(size, tuple) else size
    if max_size > 128:
        # Numba scales well for large filter sizes:
        output = numba_cpu_uniform_filter(image, size=size, mode="nearest")
        if printout_choice:
            lprint(f"Computed filter of size: {size} with Numba")
    else:
        # Scipy parallel is more efficient for small filter sizes:
        output = parallel_uniform_filter(image, size=size, mode="nearest")
        if printout_choice:
            lprint(f"Computed filter of size: {size} with parallel scipy")

    return output



# flake8: noqa
import numpy
from skimage.metrics import peak_signal_noise_ratio, structural_similarity
from skimage.util import random_noise

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, camera
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.highpass import HighpassTransform
from aydin.it.transforms.range import RangeTransform
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo_high_pass_real():

    Log.override_test_exclusion = True
    Log.enable_output = True

    # #Salt&Pepper:
    # image = normalise(camera().astype(numpy.float32))
    # noisy = image
    # noisy = random_noise(noisy, mode="s&p", amount=0.1)

    # #Gaussian:
    # image = normalise(camera().astype(numpy.float32))
    # noisy = image
    # noisy = random_noise(noisy, mode="gaussian", var=0.001)

    # Poisson:
    image = camera() // 5

    noisy = image
    noisy = random_noise(noisy, mode="poisson", clip=False).astype(numpy.float32)
    noisy = normalise(noisy)

    ac = HighpassTransform()

    preprocessed = ac.preprocess(noisy)
    postprocessed = ac.postprocess(preprocessed)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(preprocessed, name='preprocessed')
        viewer.add_image(postprocessed, name='postprocessed')

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )
    regressor = CBRegressor(patience=16)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.add_transform(RangeTransform())
    it.train(noisy)
    denoised_without_preprocessing = it.translate(noisy)

    image = numpy.clip(image, 0, 1)
    denoised_without_preprocessing = numpy.clip(denoised_without_preprocessing, 0, 1)

    psnr_denoised_without = peak_signal_noise_ratio(
        image, denoised_without_preprocessing
    )
    ssim_denoised_without = structural_similarity(image, denoised_without_preprocessing)

    results = []
    results.append(
        f"ssim_denoised_without={ssim_denoised_without}, psnr_denoised_without={psnr_denoised_without}"
    )

    denoised_images_with_preprocessing = []
    low_pass_images = []
    preprocessed_images = []

    for i in range(10):
        scale = 1 + i
        transform = HighpassTransform(scale, median_filtering=True)
        low_pass_images.append(transform._low_pass_filtering(noisy))
        preprocessed_images.append(transform.preprocess(noisy))

        it.clear_transforms()
        it.add_transform(RangeTransform())
        it.add_transform(transform)
        it.train(noisy)
        denoised_image_with_preprocessing = numpy.clip(it.translate(noisy), 0, 1)
        denoised_images_with_preprocessing.append(denoised_image_with_preprocessing)

        psnr_denoised_with = peak_signal_noise_ratio(
            image, denoised_image_with_preprocessing
        )
        ssim_denoised_with = structural_similarity(
            image, denoised_image_with_preprocessing
        )
        results.append(
            f"scale={scale}, ssim_denoised_with={ssim_denoised_with}, psnr_denoised_with={psnr_denoised_with}"
        )

    for result in results:
        print(result)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(preprocessed, name='preprocessed')
        viewer.add_image(postprocessed, name='postprocessed')
        viewer.add_image(
            denoised_without_preprocessing, name='denoised_without_preprocessing'
        )

        viewer.add_image(
            numpy.stack(denoised_images_with_preprocessing),
            name='denoised_images_with_preprocessing',
        )

        viewer.add_image(numpy.stack(preprocessed_images), name='preprocessed_images')

        viewer.add_image(numpy.stack(low_pass_images), name='low_pass_images')


demo_high_pass_real()



import numpy

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import camera


def test_standard_features_type_conservation():

    _run_type_conservation_test(numpy.float16)
    _run_type_conservation_test(numpy.float32)
    _run_type_conservation_test(numpy.float64)
    _run_type_conservation_test(numpy.uint8)
    _run_type_conservation_test(numpy.uint16)
    _run_type_conservation_test(numpy.uint32)
    _run_type_conservation_test(numpy.uint64)


def _run_type_conservation_test(dtype, required_dtype=None):
    image = camera().astype(dtype)
    # feature generator requires images in 'standard' form: BCTZYX
    image = image[numpy.newaxis, numpy.newaxis, ...]
    feature_gen = StandardFeatureGenerator(dtype=required_dtype)
    features = feature_gen.compute(image, exclude_center_feature=True)

    # float16 is a nasty case:
    if image.dtype == numpy.float16:
        assert features.dtype == numpy.float32
    else:
        assert features.dtype == image.dtype



import time
from os.path import join
import numpy
import pytest
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise
from aydin.io.folders import get_temp_folder
from aydin.it.base import ImageTranslatorBase
from aydin.it.classic import ImageDenoiserClassic


@pytest.mark.heavy
def test_saveload_bilateral():
    saveload("bilateral", min_psnr=16, min_ssim=0.30)


def test_saveload_gaussian():
    saveload("gaussian", min_psnr=20, min_ssim=0.71)


@pytest.mark.heavy
def test_saveload_gm():
    saveload("gm", min_psnr=20, min_ssim=0.65)


def test_saveload_lipschitz():
    saveload("lipschitz", min_psnr=12, min_ssim=0.16)


# @pytest.mark.heavy
def test_saveload_pca():
    saveload("pca", min_psnr=20, min_ssim=0.60)


@pytest.mark.heavy
def test_saveload_tv():
    saveload("tv", min_psnr=20, min_ssim=0.73)


@pytest.mark.heavy
def test_saveload_wavelet():
    saveload("wavelet", min_psnr=17, min_ssim=0.40)


def saveload(method, min_psnr=22, min_ssim=0.75):
    image = normalise(camera().astype(numpy.float32))
    noisy = add_noise(image)

    it = ImageDenoiserClassic(method=method)

    it.train(noisy, noisy)

    temp_file = join(get_temp_folder(), "test_it_saveload" + str(time.time()))
    print(temp_file)

    it.save(temp_file)
    del it

    loaded_it = ImageTranslatorBase.load(temp_file)

    denoised = loaded_it.translate(noisy)

    denoised = denoised.clip(0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy

    # if the line below fails, then the parameters of the image the lgbm regressohave   been broken.
    # do not change the number below, but instead, fix the problem -- most likely a parameter.

    assert psnr_denoised > min_psnr and ssim_denoised > min_ssim



"""
Define functions needed for the demos.
"""

import numpy as np
from scipy.fftpack import fft2, ifft2, fftshift, ifftshift
from scipy.signal import fftconvolve
from bm3d import gaussian_kernel


def get_psnr(y_est: np.ndarray, y_ref: np.ndarray) -> float:
    """
    Return PSNR value for y_est and y_ref presuming the noise-free maximum is 1.
    :param y_est: Estimate array
    :param y_ref: Noise-free reference
    :return: PSNR value
    """
    return 10 * np.log10(1 / np.mean(((y_est - y_ref).ravel()) ** 2))


def get_cropped_psnr(y_est: np.ndarray, y_ref: np.ndarray, crop: tuple) -> float:
    """
    Return PSNR value for y_est and y_ref presuming the noise-free maximum is 1.
    Crop the images before calculating the value by crop.
    :param y_est: Estimate array
    :param y_ref: Noise-free reference
    :param crop: Tuple of crop-x and crop-y from both stides
    :return: PSNR value
    """
    return get_psnr(
        np.atleast_3d(y_est)[crop[0] : -crop[0], crop[1] : -crop[1], :],
        np.atleast_3d(y_ref)[crop[0] : -crop[0], crop[1] : -crop[1], :],
    )


def get_experiment_kernel(
    noise_type: str, noise_var: float, sz: tuple = np.array((101, 101))
):
    """
    Get kernel for generating noise from specific experiment from the paper.
    :param noise_type: Noise type string, g[0-4](w|)
    :param noise_var: noise variance
    :param sz: size of image, used only for g4 and g4w
    :return: experiment kernel with the l2-norm equal to variance
    """
    # if noiseType == gw / g0
    kernel = np.array([[1]])
    noise_types = ['gw', 'g0', 'g1', 'g2', 'g3', 'g4', 'g1w', 'g2w', 'g3w', 'g4w']
    if noise_type not in noise_types:
        raise ValueError("Noise type must be one of " + str(noise_types))

    if noise_type != "g4" and noise_type != "g4w":
        # Crop this size of kernel when generating,
        # unless pink noise, in which
        # if noiseType == we want to use the full image size
        sz = np.array([101, 101])
    else:
        sz = np.array(sz)

    # Sizes for meshgrids
    sz2 = -(1 - (sz % 2)) * 1 + np.floor(sz / 2)
    sz1 = np.floor(sz / 2)
    uu, vv = np.meshgrid(
        [i for i in range(-int(sz1[0]), int(sz2[0]) + 1)],
        [i for i in range(-int(sz1[1]), int(sz2[1]) + 1)],
    )

    beta = 0.8

    if noise_type[0:2] == 'g1':
        # Horizontal line
        kernel = np.atleast_2d(16 - abs(np.linspace(1, 31, 31) - 16))

    elif noise_type[0:2] == 'g2':
        # Circular repeating pattern
        scale = 1
        dist = uu ** 2 + vv ** 2
        kernel = np.cos(np.sqrt(dist) / scale) * gaussian_kernel((sz[0], sz[1]), 10)

    elif noise_type[0:2] == 'g3':
        # Diagonal line pattern kernel
        scale = 1
        kernel = np.cos((uu + vv) / scale) * gaussian_kernel((sz[0], sz[1]), 10)

    elif noise_type[0:2] == 'g4':
        # Pink noise
        dist = uu ** 2 + vv ** 2
        n = sz[0] * sz[1]
        spec = np.sqrt((np.sqrt(n) * 1e-2) / (np.sqrt(dist) + np.sqrt(n) * 1e-2))
        kernel = fftshift(ifft2(ifftshift(spec)))

    else:  # gw and g0 are white
        beta = 0

    # -- Noise with additional white component --

    if len(noise_type) > 2 and noise_type[2] == 'w':
        kernel = kernel / np.sqrt(np.sum(kernel ** 2))
        kalpha = np.sqrt((1 - beta) + beta * abs(fft2(kernel, (sz[0], sz[1]))) ** 2)
        kernel = fftshift(ifft2(kalpha))

    kernel = np.real(kernel)
    # Correct variance
    kernel = kernel / np.sqrt(np.sum(kernel ** 2)) * np.sqrt(noise_var)

    return kernel


def get_experiment_noise(
    noise_type: str, noise_var: float, realization: int, sz: tuple
) -> (np.ndarray, np.ndarray, np.ndarray):
    """
    Generate noise for experiment with specified kernel, variance, seed and size.
    Return noise and relevant parameters.
    The generated noise is non-circular.
    :param noise_type: Noise type, see get_experiment_kernel for list of accepted types.
    :param noise_var: Noise variance of the resulting noise
    :param realization: Seed for the noise realization
    :param sz: image size -> size of resulting noise
    :return: noise, PSD, and kernel
    """
    np.random.seed(realization)

    # Get pre-specified kernel
    kernel = get_experiment_kernel(noise_type, noise_var, sz)

    # Create noisy image
    half_kernel = np.ceil(np.array(kernel.shape) / 2)

    if len(sz) == 3 and half_kernel.size == 2:
        half_kernel = [half_kernel[0], half_kernel[1], 0]
        kernel = np.atleast_3d(kernel)

    half_kernel = np.array(half_kernel, dtype=int)

    # Crop edges
    noise = fftconvolve(
        np.random.normal(size=(sz + 2 * half_kernel)), kernel, mode='same'
    )
    noise = np.atleast_3d(noise)[
        half_kernel[0] : -half_kernel[0], half_kernel[1] : -half_kernel[1], :
    ]

    psd = abs(fft2(kernel, (sz[0], sz[1]), axes=(0, 1))) ** 2 * sz[0] * sz[1]

    return noise, psd, kernel



import numpy

# from numpy.typing import ArrayLike
from numpy import sort
from scipy.ndimage import uniform_filter

from aydin.it.classic_denoisers.lipschitz import denoise_lipschitz
from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class SaltPepperTransform(ImageTransformBase):
    """Salt And Pepper Correction

    Detectors such as cameras have 'broken' pixels that blink, are very dim,
    or very bright. Other phenomena cause voxels to have very different
    values from their neighbors, this is often called 'impulse' or
    'salt-and-pepper' noise. While self-supervised denoising can solve many of
    these issues, there is no reason to not to alleviate the task, especially
    when there are simple and fast approaches that can tackle this kind of
    noise. This preprocessing replaces voxels with the median of its
    neighbors if the voxel value is too different from its neighbors. This
    difference is proportional to the local second-derivative of the image.
    Increase the threshold parameter to tolerate more variation, decrease it
    to be more aggressive in removing salt & pepper noise. The algorithm is
    iterative, starting with the most offending pixels, until no pixels are
    corrected. You can set the max proportion of pixels that are allowed to
    be corrected if you can give a good estimate for that.
    """

    preprocess_description = (
        "Salt and pepper pixels correction" + ImageTransformBase.preprocess_description
    )
    postprocess_description = "Not supported (why would anyone want to do that? ☺)"
    postprocess_supported = False
    postprocess_recommended = False

    def __init__(
        self,
        fix_repeated: bool = True,
        max_repeated: int = 4,
        fix_lipschitz: bool = True,
        lipschitz: float = 0.1,
        percentile: float = 0.01,
        num_iterations: int = 64,
        priority: float = 0.08,
        **kwargs,
    ):

        """
        Constructs a Salt And Pepper Transform

        Parameters
        ----------

        fix_repeated: bool
             Removes Salt & pepper by finding highly repeated values.
             These values are then considered as erroneous and are fixed
             by interpolation.

        max_repeated: int
            Max number of repeated values to fix.

        fix_lipschitz: bool
            Removes Salt & pepper by enforcing Lipschitz continuity.

        lipschitz : float
            Lipschitz threshold. Increase to tolerate more variation, decrease to be
            more aggressive in removing impulse/salt&pepper noise.

        percentile : float
            Percentile value used to determine the threshold
            for choosing the worst offending voxels per iteration
            according to the Lipschitz threshold.

        num_iterations : int
            Number of iterations for enforcing Lipschitz continuity.

        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)

        self.fix_lipschitz = fix_lipschitz
        self.num_iterations = num_iterations
        self.correction_percentile = percentile
        self.lipschitz = lipschitz
        self.fix_repeated = fix_repeated
        self.max_repeated = max_repeated

        self._original_dtype = None

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_original_dtype']
        return state

    def __str__(self):
        return (
            f'{type(self).__name__} (fix_lipschitz={self.fix_lipschitz},'
            f' num_iterations={self.num_iterations},'
            f' correction_percentile={self.correction_percentile},'
            f' lipschitz={self.lipschitz},'
            f' fix_repeated={self.fix_repeated},'
            f' max_repeated={self.max_repeated} )'
        )

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):

        with lsection(
            f"Broken Pixels Correction for array of shape: {array.shape} and dtype: {array.dtype}:"
        ):
            # We save the original dtype:
            self._original_dtype = array.dtype

            # If needed, we convert to float32:
            array = array.astype(dtype=numpy.float32, copy=True)

            # First we look at over represented voxel values -- a sign of problematic voxels,
            # and try to fix them:
            if self.fix_repeated:
                array = self._repeated_value_method(array)

            # Then we enforce Lipschitz continuity:
            if self.fix_lipschitz:
                array = self._lipschitz_method(array)

            return array

    def postprocess(self, array):
        # undoing this transform is unpractical and unlikely to be usefull
        array = array.astype(self._original_dtype, copy=False)
        return array

    def _repeated_value_method(self, array):
        with lsection(
            "Correcting for wrong pixels values using the 'repeated-value' approach:"
        ):
            unique, counts = numpy.unique(array, return_counts=True)

            # How many unique values in image?
            num_unique_values = unique.size
            lprint(f"Number of unique values in image: {num_unique_values}.")

            # Most occuring value
            most_occuring_value = unique[numpy.argmax(counts)]
            highest_count = numpy.max(counts)
            lprint(
                f"Most occurring value in array: {most_occuring_value}, {highest_count} times."
            )

            # Assuming a uniform distribution we would expect each value to be used at most:
            average_count = array.size // num_unique_values
            lprint(
                f"Average number of occurences of a value assuming uniform distribution: {average_count}"
            )

            # We fix at most n over-represented values:
            selected_counts = sort(counts.flatten())

            # First we ignore counts below a certain thresholds:
            selected_counts = selected_counts[selected_counts > average_count]

            # use Otsu split to clean up remaining values:
            mask = _otsu_split(selected_counts)
            selected_counts = selected_counts[mask]

            # Maximum number of repeated values to remove:
            n = self.max_repeated
            n = min(n, len(selected_counts))
            max_tolerated_count = selected_counts[-n]
            lprint(f"Maximum tolerated count per value: {max_tolerated_count}.")

            # If a voxel value appears over more than 0.1% of voxels, then it is a problematic value:
            problematic_counts_mask = counts > max_tolerated_count
            problematic_counts = counts[problematic_counts_mask]
            problematic_values = unique[problematic_counts_mask]

            lprint(f"Problematic values: {list(problematic_values)}.")
            lprint(f"Problematic counts: {list(problematic_counts)}.")

            # We construct the mask of good values:
            good_values_mask = numpy.ones_like(array, dtype=numpy.bool_)
            for problematic_value in problematic_values:
                good_values_mask &= array != problematic_value

            with lsection(f"Correcting voxels with values: {problematic_values}."):

                # We save the good values (copy!):
                good_values = array[good_values_mask].copy()

                # We compute the number of iterations:
                num_bad_values = array.size - len(good_values)
                num_iterations = 16 * int(
                    (array.size / num_bad_values) ** (1.0 / array.ndim)
                )

                # We solve the harmonic equation:
                for i in range(num_iterations):
                    lprint(f"Iteration {i}")
                    # We compute the median:
                    array = uniform_filter(array, size=3)
                    # We use the median to correct pixels:
                    array[good_values_mask] = good_values

                # count number of corrections for this round:
                num_corrections = numpy.sum(mask)
                lprint(f"Number of corrections: {num_corrections}.")

        return array

    def _lipschitz_method(self, array):
        # Iterations:
        with lsection(
            "Correcting for wrong pixels values using the Lipschitz approach:"
        ):
            array = denoise_lipschitz(
                array,
                lipschitz=self.lipschitz,
                percentile=self.correction_percentile,
                max_num_iterations=self.num_iterations,
            )

            return array

            # OLD METHOD KEEP!
            # for i in range(self.num_iterations):
            #     lprint(f"Iteration {i}")
            #
            #     # Compute median:
            #     median = median_filter(array, size=3)
            #
            #     # We scale the lipschitz threshold to the image std at '3 sigma' :
            #     lipschitz = self.lipschitz * 3 * median.std()
            #
            #     # We compute the 'error':
            #     median, error = self._compute_error(
            #         array, median=median, lipschitz=lipschitz
            #     )
            #
            #     # We compute the threshold on the basis of the errors,
            #     # we first tackle the most offending voxels:
            #     threshold = numpy.percentile(
            #         error, q=100 * (1 - self.correction_percentile)
            #     )
            #
            #     # We compute the mask:
            #     mask = error > threshold
            #
            #     # count number of corrections for this round:
            #     num_corrections = numpy.sum(mask)
            #     lprint(f"Number of corrections: {num_corrections}")
            #
            #     # if no corrections made we stop iterating:
            #     if num_corrections == 0:
            #         break
            #
            #     # We keep track of the proportion of voxels corrected:
            #     proportion = (
            #         num_corrections + total_number_of_corrections
            #     ) / array.size
            #     lprint(
            #         f"Proportion of corrected pixels: {int(proportion * 100)}% (up to now), versus maximum: {int(self.max_proportion_corrected * 100)}%) "
            #     )
            #
            #     # If too many voxels have been corrected we stop:
            #     if proportion > self.max_proportion_corrected:
            #         break
            #
            #     # We use the median to correct pixels:
            #     array[mask] = median[mask]
            #
            #     # increment total number of corrections:
            #     total_number_of_corrections += num_corrections

    def _compute_error(self, array, median, lipschitz):
        # we compute the error map:
        error = median.copy()
        error -= array
        numpy.abs(error, out=error)
        numpy.maximum(error, lipschitz, out=error)
        error -= lipschitz
        return median, error


def _otsu_split(array):
    # Flatten array:
    shape = array.shape
    array = array.reshape(-1)

    mean_weigth = 1.0 / array.size
    his, bins = numpy.histogram(array, bins='auto', density=True)
    final_thresh = -1
    final_value = -1
    for i in range(1, len(bins) - 1):
        Wb = numpy.sum(his[:i]) * mean_weigth
        Wf = numpy.sum(his[i:]) * mean_weigth

        mub = numpy.mean(his[:i])
        muf = numpy.mean(his[i:])

        value = Wb * Wf * (mub - muf) ** 2

        # print("Wb", Wb, "Wf", Wf)
        # print("t", i, "value", value)

        if value > final_value:
            final_thresh = 0.5 * (bins[i] + bins[i + 1])
            final_value = value

    mask = array > final_thresh

    mask = mask.reshape(shape)

    return mask



import gc
import random
from os.path import join, exists
import numpy
import psutil
from tensorflow.python.eager.context import device
from tensorflow.python.keras.optimizer_v2.adam import Adam
from tensorflow.python.keras.saving.model_config import model_from_json

from aydin.io.folders import get_temp_folder
from aydin.regression.nn_utils.callbacks import (
    NNCallback,
    EarlyStopping,
    ReduceLROnPlateau,
    ModelCheckpoint,
)
from aydin.regression.nn_utils.models import feed_forward
from aydin.regression.base import RegressorBase
from aydin.util.log.log import lsection, lprint
from aydin.util.tf.device import get_best_device_name


class NNRegressor(RegressorBase):
    """standard perceptron-like neural network regressor."""

    device_max_mem = psutil.virtual_memory().total

    def __init__(
        self,
        max_epochs: int = 1024,
        learning_rate: float = 0.001,
        patience: int = 10,
        depth: int = 6,
        loss: str = 'l1',
    ):
        """

        Parameters
        ----------
        max_epochs : int
            Maximum number of epochs allowed
        learning_rate : float
            Learning rate
        patience : int
            Number of epochs required for early stopping
        depth : int
            Depth of the model
        loss : str
            Type of loss to be used

        """
        super().__init__()

        self.max_epochs = max_epochs
        self.learning_rate = learning_rate
        self.patience = patience
        self.depth = depth

        loss = 'mae' if loss.lower() == 'l1' else loss
        loss = 'mse' if loss.lower() == 'l2' else loss
        self.loss = loss

        with lsection("NN Regressor"):
            lprint("with no arguments")  # TODO: fix these logs

    def _fit(
        self, x_train, y_train, x_valid=None, y_valid=None, regressor_callback=None
    ):
        """Fits function y=f(x) given training pairs (x_train, y_train).
        Stops when performance stops improving on the test dataset: (x_test, y_test).
        """

        with lsection("NN Regressor fitting:"):

            with device(get_best_device_name()):
                # First we make sure that the arrays are of a type supported:
                def assert_type(array):
                    assert (
                        (array.dtype == numpy.float64)
                        or (array.dtype == numpy.float32)
                        or (array.dtype == numpy.float16)
                        or (array.dtype == numpy.uint16)
                        or (array.dtype == numpy.uint8)
                    )

                # Do we have a validation dataset?
                has_valid_dataset = x_valid is not None and y_valid is not None

                assert_type(x_train)
                assert_type(y_train)
                if has_valid_dataset:
                    assert_type(x_valid)
                    assert_type(y_valid)

                    # Types have to be consistent between train and valid sets:
                    assert x_train.dtype is x_valid.dtype
                    assert y_train.dtype is y_valid.dtype

                # In case the y dtype does not match the x dtype, we rescale and cast y:
                if numpy.issubdtype(x_train.dtype, numpy.integer) and numpy.issubdtype(
                    y_train.dtype, numpy.floating
                ):

                    # We remember the original type of y:
                    original_y_dtype = y_train.dtype

                    if x_train.dtype == numpy.uint8:
                        y_train *= 255
                        y_train = y_train.astype(numpy.uint8, copy=False)
                        if has_valid_dataset:
                            y_valid *= 255
                            y_valid = y_valid.astype(numpy.uint8, copy=False)
                        original_y_scale = 1 / 255.0

                    elif x_train.dtype == numpy.uint16:
                        y_train *= 255 * 255
                        y_train = y_train.astype(numpy.uint16, copy=False)
                        if has_valid_dataset:
                            y_valid *= 255 * 255
                            y_valid = y_valid.astype(numpy.uint16, copy=False)
                        original_y_scale = 1 / (255.0 * 255.0)
                else:
                    original_y_dtype = None
                    original_y_scale = None

                # Get the number of entries and features from the array shape:
                nb_data_points = x_train.shape[0]
                num_features = x_train.shape[-1]

                lprint(f"Number of data points : {nb_data_points}")
                if has_valid_dataset:
                    lprint(f"Number of validation data points: {x_valid.shape[0]}")
                lprint(f"Number of features per data point: {num_features}")

                # Shapes of both x and y arrays:
                x_shape = (-1, num_features)
                y_shape = (-1, 1)

                # Learning rate and decay:
                learning_rate = self.learning_rate
                learning_rate_decay = 0.1 * self.learning_rate
                lprint(f"Learning rate: {learning_rate}")
                lprint(f"Learning rate decay: {learning_rate_decay}")

                # Weight decay and noise:
                weight_decay = 0.01 * self.learning_rate
                noise = 0.1 * self.learning_rate
                lprint(f"Weight decay: {weight_decay}")
                lprint(f"Added noise: {noise}")

                # Initialise model if not done yet:
                model = feed_forward(
                    num_features,
                    depth=self.depth,
                    weight_decay=weight_decay,
                    noise=noise,
                )
                opt = Adam(lr=learning_rate, decay=learning_rate_decay)
                model.compile(optimizer=opt, loss=self.loss)

                lprint(f"Number of parameters in model: {model.count_params()}")

                # Reshape arrays:
                x_train = x_train.reshape(x_shape)
                y_train = y_train.reshape(y_shape)

                if x_valid is not None and y_valid is not None:
                    x_valid = x_valid.reshape(x_shape)
                    y_valid = y_valid.reshape(y_shape)

                batch_size = 1024
                lprint(f"Keras batch size for training: {batch_size}")

                # Effective number of epochs:
                effective_number_of_epochs = self.max_epochs
                lprint(f"Effective max number of epochs: {effective_number_of_epochs}")

                # Early stopping patience:
                early_stopping_patience = self.patience
                lprint(f"Early stopping patience: {early_stopping_patience}")

                # Effective LR patience:
                effective_lr_patience = max(1, self.patience // 2)
                lprint(f"Effective LR patience: {effective_lr_patience}")

                # Here is the list of callbacks:
                callbacks = []

                # Set upstream callback:
                keras_callback = NNCallback(regressor_callback)

                # Early stopping callback:
                early_stopping = EarlyStopping(
                    self,
                    monitor='val_loss',
                    min_delta=min(0.000001, 0.1 * self.learning_rate),
                    patience=early_stopping_patience,
                    mode='auto',
                    restore_best_weights=True,
                )

                # Reduce LR on plateau:
                reduce_learning_rate = ReduceLROnPlateau(
                    monitor='val_loss',
                    factor=0.5,
                    verbose=1,
                    patience=effective_lr_patience,
                    mode='auto',
                    min_lr=0.0001 * self.learning_rate,
                )

                model_file_path = join(
                    get_temp_folder(),
                    f"aydin_nn_keras_model_file_{random.randint(0, 1e16)}.hdf5",
                )
                checkpoint = ModelCheckpoint(
                    model_file_path, monitor='val_loss', verbose=1, save_best_only=True
                )

                # Add callbacks to the list:
                callbacks.append(keras_callback)
                callbacks.append(early_stopping)
                callbacks.append(reduce_learning_rate)
                callbacks.append(checkpoint)

                # x_train = x_train.astype(numpy.float64)
                # y_train = y_train.astype(numpy.float64)
                # x_valid = x_valid.astype(numpy.float64)
                # y_valid = y_valid.astype(numpy.float64)

                # Training happens here:
                with lsection("NN regressor fitting now:"):
                    train_history = model.fit(
                        x_train,
                        y_train,
                        validation_data=(x_valid, y_valid)
                        if (x_valid is not None and y_valid is not None)
                        else None,
                        epochs=effective_number_of_epochs,
                        batch_size=min(batch_size, nb_data_points),
                        shuffle=True,
                        verbose=0,  # 0 if is_batch else 1,
                        callbacks=callbacks,
                    )
                    lprint("NN regressor fitting done.")

                del x_train
                del y_train

                # Reload the best weights:
                if exists(model_file_path):
                    lprint("Loading best model to date.")
                    model.load_weights(model_file_path)

                # loss_history = train_history.history['loss']
                # lprint(f"Loss history after training: {loss_history}")

                if 'val_loss' in train_history.history:
                    self.last_val_loss = train_history.history['val_loss'][0]

                loss_history = {
                    'training': train_history.history['loss'],
                    'validation': train_history.history['val_loss'],
                }

                gc.collect()

                return _NNModel(model, original_y_dtype, original_y_scale, loss_history)


class _NNModel:
    def __init__(self, model, original_y_dtype, original_y_scale, loss_history):
        self.model = model
        self.original_y_dtype = original_y_dtype
        self.original_y_scale = original_y_scale
        self.loss_history = loss_history

    def _save_internals(self, path: str):

        if self.model is not None:
            # serialize model to JSON:
            keras_model_file = join(path, 'keras_model.txt')
            model_json = self.model.to_json()
            with open(keras_model_file, "w") as json_file:
                json_file.write(model_json)

            # serialize weights to HDF5:
            keras_model_file = join(path, 'keras_weights.txt')
            self.model.save_weights(keras_model_file)

    def _load_internals(self, path: str):
        # load JSON and create model:
        keras_model_file = join(path, 'keras_model.txt')
        with open(keras_model_file, 'r') as json_file:
            loaded_model_json = json_file.read()
            self.model = model_from_json(loaded_model_json)
        # load weights into new model:
        keras_model_file = join(path, 'keras_weights.txt')
        self.model.load_weights(keras_model_file).expect_partial()

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['model']
        return state

    def predict(self, x):
        """Predicts y given x by applying the learned function f: y=f(x)

        Parameters
        ----------
        x

        Returns
        -------

        """
        with lsection("NN Regressor prediction:"):

            with device(get_best_device_name()):
                lprint(f"Number of data points             : {x.shape[0]}")
                lprint(f"Number of features per data points: {x.shape[-1]}")

                # Number of features:
                num_of_features = x.shape[-1]

                # We check that we get the right number of features.
                # If not, most likely the batch_dims are set wrong...
                assert num_of_features == x.shape[-1]

                # How much memory is available in GPU:
                max_gpu_mem_in_bytes = NNRegressor.device_max_mem

                # We limit ourselves to using only a quarter of GPU memory:
                max_number_of_floats = (max_gpu_mem_in_bytes // 4) // 4

                # Max size of batch:
                max_gpu_batch_size = max_number_of_floats / num_of_features

                # Batch size taking all this into account:
                batch_size = max(1, min(max_gpu_batch_size, x.shape[0] // 256))

                # Heuristic threshold here obtained by inspecting batch size per GPU memory
                # Basically ensures ratio of 700000 batch size per 12GBs of GPU memory
                batch_size = min(
                    batch_size, (700000 * max_gpu_mem_in_bytes) // 12884901888
                )

                lprint(f"Batch size: {batch_size}")
                lprint(f"Predicting. features shape = {x.shape}")

                lprint("NN regressor predicting now...")
                yp = self.model.predict(x, batch_size=batch_size)
                lprint("NN regressor predicting done!")

                # We cast back yp to the correct type and range:
                if self.original_y_dtype is not None:
                    yp = yp.astype(self.original_y_dtype)
                    yp *= self.original_y_scale

                return yp



import time
from os.path import join

import numpy
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import add_noise, normalise
from aydin.io.folders import get_temp_folder
from aydin.regression.base import RegressorBase
from aydin.regression.cb import CBRegressor
from aydin.regression.lgbm import LGBMRegressor
from aydin.regression.linear import LinearRegressor
from aydin.regression.nn import NNRegressor
from aydin.regression.rf import RFRegressor
from aydin.regression.sv import SVRegressor


def test_linear_saveload():
    saveload(LinearRegressor(), min_ssim=0.70)


def test_rf_saveload():
    saveload(RFRegressor(), min_ssim=0.70)


def test_svr_saveload():
    saveload(SVRegressor(), min_ssim=0.70)


def test_gbm_saveload():
    saveload(LGBMRegressor(max_num_estimators=128), min_ssim=0.73)


def test_cb_saveload():
    saveload(CBRegressor(max_num_estimators=128, min_num_estimators=64), min_ssim=0.72)


def test_nn_saveload():
    saveload(NNRegressor(max_epochs=10), min_ssim=0.71)


def saveload(regressor, min_ssim=0.80):

    image = camera().astype(numpy.float32)
    image = normalise(image)
    noisy = add_noise(image)

    # feature generator requires images in 'standard' form: BCTZYX
    noisy = noisy[numpy.newaxis, numpy.newaxis, ...]

    generator = StandardFeatureGenerator()

    features = generator.compute(noisy, exclude_center_value=True)

    x = features.reshape(-1, features.shape[-1])
    y = noisy.reshape(-1)

    regressor.fit(x, y)

    temp_file = join(get_temp_folder(), "test_reg_saveload.json" + str(time.time()))
    regressor.save(temp_file)

    del regressor

    loaded_regressor = RegressorBase.load(temp_file)

    yp = loaded_regressor.predict(x)

    denoised = yp.reshape(image.shape)

    # print(numpy.min(denoised), numpy.min(image))
    # print(numpy.max(denoised), numpy.max(image))
    denoised = denoised.clip(0, 1)
    ssim_value = ssim(denoised, image)
    psnr_value = psnr(denoised, image)

    print("denoised", psnr_value, ssim_value)

    assert ssim_value > min_ssim



import torch.nn as nn
import torch.nn.functional as F


class FeedForward(nn.Module):
    def __init__(
        self, n_input_channel=1, n_output_channel=1, depth=8, nic=8, kernel_size=3
    ):
        super().__init__()

        self.convs = []

        for i in range(0, depth - 1):
            in_channels = n_input_channel if i == 0 else nic
            conv = nn.Conv2d(
                in_channels,
                nic,
                kernel_size=kernel_size,
                padding=(kernel_size - 1) // 2,
            )
            self.convs.append(conv)

        self.convs = nn.ModuleList(self.convs)
        self.final_conv = nn.Conv2d(nic, n_output_channel, kernel_size=1, padding=0)

    def forward(self, x0):

        x = x0

        xn = []
        for conv in self.convs:
            x = conv(x)
            x = F.leaky_relu(x, negative_slope=0.01)
            xn.append(x)

        y = xn[0]
        s = 1
        for x in xn[1:]:
            y = y + s * x
            # s*=0.5

        y = self.final_conv(y)

        return y



from aydin.io.datasets import cropped_newyork
from aydin.it.classic_denoisers.bilateral import denoise_bilateral

from aydin.it.classic_denoisers.demo.demo_2D_bilateral import demo_bilateral
from aydin.it.classic_denoisers.test.util_test_nd import check_nd


def est_bilateral():
    # TODO: bilateral filter in scikit image seems broken
    assert demo_bilateral(cropped_newyork(), display=False) >= 0.40 - 0.1


def test_bilateral_nd():
    check_nd(denoise_bilateral)



from qtpy.QtWidgets import QHBoxLayout, QFrame, QWidget
from qtpy.QtCore import Qt


class QHorizontalLineBreakWidget(QWidget):
    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.parent = parent

        self.layout = QHBoxLayout()

        self.horizontal_line_break = QFrame(
            frameShape=QFrame.HLine, frameShadow=QFrame.Sunken
        )
        self.layout.addWidget(self.horizontal_line_break)
        self.layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.layout)



# flake8: noqa
import time

import napari
import numpy
from skimage.data import camera
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim
from skimage.util import random_noise

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.range import RangeTransform
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True

    image = camera()

    intensity = 5
    numpy.random.seed(0)
    noisy = image.astype(numpy.float32)
    noisy = rescale_intensity(noisy, in_range='image', out_range=(0, 1))
    noisy = numpy.random.poisson(noisy * intensity) / intensity
    noisy = random_noise(noisy, mode='gaussian', var=0.01, seed=0)
    noisy *= 255
    noisy = noisy.astype(numpy.uint16)

    # Both image and noisy are uint32 within the [0, 255] range...
    # Builtin normalisation should do the job...

    generator = StandardFeatureGenerator()
    regressor = LGBMRegressor()

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    it.transforms_list.append(RangeTransform())

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    image = image.astype(numpy.float32)
    noisy = noisy.astype(numpy.float32)
    denoised = denoised.astype(numpy.float32)

    image = rescale_intensity(image, in_range='image', out_range=(0, 1))
    noisy = rescale_intensity(noisy, in_range='image', out_range=(0, 1))
    denoised = rescale_intensity(denoised, in_range='image', out_range=(0, 1))

    print("noisy", psnr(image, noisy), ssim(noisy, image))
    print("denoised", psnr(image, denoised), ssim(denoised, image))

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(denoised, name='denoised')


demo()



import torch
import torch.nn as nn


class Masking(nn.Module):
    def __init__(self, module):
        super().__init__()

        self.module: nn.Module = module
        self.density = 0.1
        self.mask = None

    def forward(self, x):
        self.mask = torch.rand_like(x) < self.density
        x = (~self.mask) * x
        x = self.module(x)
        return x

    def get_mask(self):
        return self.mask

    def trainable_parameters(self):
        return self.module.trainable_parameters()



from qtpy.QtCore import Qt
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QWidget, QHBoxLayout, QLabel

from aydin.gui._qt.custom_widgets.vertical_line_break_widget import (
    QVerticalLineBreakWidget,
)


class QReadMoreLessLabel(QWidget):
    def __init__(self, parent, text):
        QWidget.__init__(self, parent)

        self.text = text
        self.readmore = False

        # Explanation text
        self.explanation_layout = QHBoxLayout()
        self.explanation_layout.setAlignment(Qt.AlignTop)
        self.mousePressEvent = self.state_toggle

        if "<moreless>" in text:
            self.readless_text, self.readmore_text = text.split("<moreless>")
            self.readmore_text = text
            self.readmore_left, self.readmore_right = self.readmore_text.split(
                "<split>"
            )
            text_left = self.readless_text + (
                "" if self.readmore_right.strip() == "" else "<b>Read more...</b>"
            )
            text_right = ""
        else:
            text_left, text_right = text.split("<split>")

        self.explanation_text_left = QLabel(text_left, self)
        self.explanation_text_left.setWordWrap(True)
        self.explanation_text_left.setTextFormat(Qt.RichText)
        self.explanation_layout.addWidget(self.explanation_text_left, 50)

        # Vertical Line Break
        self.vertical_break = QVerticalLineBreakWidget(self)
        self.explanation_layout.addWidget(self.vertical_break)

        self.explanation_text_right = QLabel(text_right, self)
        self.explanation_text_right.setWordWrap(True)
        self.explanation_text_right.setTextFormat(Qt.RichText)
        self.explanation_layout.addWidget(self.explanation_text_right, 50)

        if self.readmore_text is not None and self.readmore_text.strip() == "":
            self.vertical_break.setHidden(True)
            self.explanation_text_right.setHidden(True)

        self.setLayout(self.explanation_layout)

    def state_toggle(self, a0: QMouseEvent) -> None:

        if self.readmore_text is not None and self.readmore_right.strip() != "":
            self.readmore = not self.readmore
            if self.readmore:
                self.explanation_text_left.setText(self.readmore_left)
                self.explanation_text_right.setText(
                    self.readmore_right + "<b>Read less...</b>"
                )
            else:
                self.explanation_text_left.setText(
                    self.readless_text + "<b>Read more...</b>"
                )
                self.explanation_text_right.setText("")



from os import access, W_OK
from os.path import exists

from aydin.io.folders import get_home_folder, get_temp_folder, get_cache_folder


def test_home_folder():

    home_folder = get_home_folder()
    print(home_folder)
    assert exists(home_folder)
    assert access(home_folder, W_OK)


def test_temp_folder():

    temp_folder = get_temp_folder()
    print(temp_folder)
    assert exists(temp_folder)
    assert access(temp_folder, W_OK)


def test_cache_folder():

    cache_folder = get_cache_folder()
    print(cache_folder)
    assert exists(cache_folder)
    assert access(cache_folder, W_OK)



import math
import time
from random import randrange
from typing import Optional

import numpy
from scipy.ndimage import sobel, laplace

from aydin.util.fast_uniform_filter.uniform_filter import uniform_filter_auto
from aydin.util.log.log import lprint


def representative_crop(
    image,
    mode: str = 'sobelmin',
    crop_size: Optional[int] = None,
    min_length: int = 16,
    smoothing_size: int = 1.0,
    equal_sides: bool = False,
    favour_odd_lengths: bool = False,
    fast_mode: bool = False,
    fast_mode_num_crops: int = 1500,
    max_time_in_seconds: float = 1,
    return_slice: bool = False,
    display_crop: bool = False,
):
    """Extract a representative crop from the image. Searches for the crop of given
    (approximate) size with highest score. The score is simply the sum of sobel
    magnitudes (~tenengrad) which is a good metric for estimating how much interesting
    content each crop contains. Empirically, this highly correlates with where
    I (Loic A. Royer) tend to look at in images.


    Parameters
    ----------
    image : ArrayLike
        Image to extract representative crop from

    mode : str
        Strategy for picking crop.

    crop_size : int
        Crop size in voxels. Default (None) is 32000.

    min_length : int
        Crop axis lengths cannot be smaller than this number.

    smoothing_size : int
        Uniform filter smoothing to achieve some crude denoising and thus
        make it a bit easier to estimate the score.

    equal_sides : bool
        When True the crop will have all its sides of equal size (square, cube, ...)

    favour_odd_lengths : bool
        If possible favours crops that have odd shape lengths.

    fast_mode: bool
        In fast mode we pick random crops instead of checking every possible strided crop.

    fast_mode_num_crops: int
        Number of crops to check in fast mode.

    max_time_in_seconds: float
        Maximum amount of time in seconds that this function should run for.
        This avoids excessive computation for very large images.

    return_slice : bool
        If True the slice is returned too:

    display_crop: bool
        Displays crop, for debugging purposes...

    Returns
    -------
    Most representative crop, and if return_slice is True the actual slice object too.

    """

    # Start time:
    start_time = time.time()

    # Number of voxels in image:
    num_voxels = image.size

    # Default number of voxels:
    if crop_size is None:
        crop_size = 32000

    # Ratio by which to crop to acheive max num voxels:
    ratio = (crop_size / num_voxels) ** (1 / image.ndim)

    if ratio >= 1:
        # If the image is small enough no point in getting a crop!
        return image

    # cropped shape:
    cropped_shape = tuple(min(max(min_length, int(s * ratio)), s) for s in image.shape)

    # Favour odd lengths if requested:
    if favour_odd_lengths:
        cropped_shape = tuple((s // 2) * 2 + 1 for s in cropped_shape)

    # We enforce equal sides if requested:
    if equal_sides:
        min_length = min(cropped_shape)
        cropped_shape = tuple((min_length,) * image.ndim)

    # range for translation:
    translation_range = tuple(s - cs for s, cs in zip(image.shape, cropped_shape))

    # grid for translations:
    translation_indices = tuple(
        max(1, int(2 * r / cs)) for r, cs in zip(translation_range, cropped_shape)
    )

    # min and max for crop value normalisation:
    image_min = None
    image_max = None

    # We loop through a number of crops and keep the one wit the best score:
    best_score = -1
    best_slice = None
    best_crop = None

    if fast_mode and image.size > 1e6:

        # We make sure that the number of crops is not too large given
        # the relative size of the crop versus whole image:
        fast_mode_num_crops = min(fast_mode_num_crops, int(image.size / crop_size))

        for index in range(fast_mode_num_crops):

            # translation:
            translation = tuple(
                (randrange(0, s - cs) if cs != s else 0)
                for s, cs in zip(image.shape, cropped_shape)
            )

            # function to get crop slice:
            def _crop_slice(translation, cropped_shape, downscale: int = 1):
                return tuple(
                    slice(t, t + s, downscale)
                    for t, s in zip(translation, cropped_shape)
                )

            # slice object for cropping:
            crop_slice = _crop_slice(
                translation, cropped_shape, 2 if image.size > 1e8 else 1
            )

            # extract crop:
            crop = image[crop_slice]

            score = evaluate_crop(
                crop=crop,
                image_min=image_min,
                image_max=image_max,
                mode=mode,
                smoothing_size=smoothing_size,
            )

            # slice object for the actual crop:
            crop_slice = _crop_slice(translation, cropped_shape, 1)

            # update best score and image:
            if score > best_score and not math.isinf(score):
                best_score = score
                best_slice = crop_slice

                # We make sure to have the full and original crop!
                best_crop = image[best_slice]

            if time.time() > start_time + max_time_in_seconds:
                lprint("Interrupting crop search because of timeout!")
                break

    else:
        for index in numpy.ndindex(translation_indices):

            # translation:
            translation = tuple(int(i * cs / 2) for i, cs in zip(index, cropped_shape))

            # slice object for cropping:
            crop_slice = tuple(
                slice(t, t + s) for t, s in zip(translation, cropped_shape)
            )

            # extract crop:
            crop = image[crop_slice]

            score = evaluate_crop(
                crop=crop,
                image_min=image_min,
                image_max=image_max,
                mode=mode,
                smoothing_size=smoothing_size,
            )

            # update best score and image:
            if score > best_score and not math.isinf(score):
                best_score = score
                best_slice = crop_slice

                # We make sure to have the full and original crop!
                best_crop = image[best_slice]

            if time.time() > start_time + max_time_in_seconds:
                lprint("Interrupting crop search because of timeout!")
                break

    if display_crop:
        smoothed_best_crop = _smoothing(best_crop, size=smoothing_size)

        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(best_crop, name='best_crop')
            viewer.add_image(smoothed_best_crop, name='smoothed_best_crop')

    if return_slice:
        return best_crop, best_slice
    else:
        return best_crop


def evaluate_crop(crop, image_min, image_max, mode, smoothing_size):
    # smooth crop as a crude denoising:
    smoothed_crop = _smoothing(crop, size=smoothing_size)
    # convert type:
    smoothed_crop = smoothed_crop.astype(dtype=numpy.float32, copy=False)

    if image_min is None or image_max is None:
        image_min = crop.min()
        image_max = crop.max()
    else:
        # Normalise crop values:
        smoothed_crop -= image_min
        smoothed_crop /= image_max
    # compute score:
    if mode == 'contrast':
        score = numpy.std(smoothed_crop)
    elif mode == 'sobel':
        score = numpy.std(_sobel_magnitude(smoothed_crop))
    elif mode == 'sobelmin':
        score = numpy.std(_sobel_minimum(smoothed_crop))
    elif mode == 'laplace':
        score = numpy.std(laplace(smoothed_crop))
    else:
        raise ValueError(f"Unknown mode: {mode}")
    return score


def _smoothing(crop, size):
    smoothed_crop = uniform_filter_auto(crop, size=1 + 2 * size, printout_choice=False)
    # smoothed_crop = gaussian_filter(crop, sigma=size)
    return smoothed_crop


def _sobel_magnitude(image):

    magnitude = numpy.zeros_like(image)
    for axis in range(image.ndim):
        magnitude += sobel(image, axis=axis) ** 2

    return numpy.sqrt(magnitude)


def _sobel_minimum(image):

    minimum = None
    for axis in range(image.ndim):
        if minimum is None:
            minimum = sobel(image, axis=axis)
        else:
            minimum = numpy.minimum(minimum, sobel(image, axis=axis))

    return minimum



import numpy as np

from aydin.nn.models.jinet import JINetModel
from aydin.nn.models.unet import UNetModel


def test_supervised_2D():
    input_array = np.zeros((1, 64, 64, 1), dtype=np.float32)
    model2d = UNetModel(
        (64, 64, 1),
        nb_unet_levels=2,
        shiftconv=False,
        supervised=True,
        spacetime_ndim=2,
    )
    result = model2d.predict(input_array)
    assert result.shape == input_array.shape
    assert result.dtype == input_array.dtype


def test_shiftconv_2D():
    input_array = np.zeros((1, 64, 64, 1), dtype=np.float32)
    model2d = UNetModel(
        (64, 64, 1),
        nb_unet_levels=2,
        shiftconv=True,
        supervised=False,
        spacetime_ndim=2,
    )
    result = model2d.predict(input_array)
    assert result.shape == input_array.shape
    assert result.dtype == input_array.dtype


def test_masking_2D():
    input_array = np.zeros((1, 64, 64, 1), dtype=np.float32)
    model2d = UNetModel(
        (64, 64, 1),
        nb_unet_levels=2,
        shiftconv=False,
        supervised=False,
        spacetime_ndim=2,
    )
    result = model2d.predict([input_array, input_array])
    assert result.shape == input_array.shape
    assert result.dtype == input_array.dtype


def test_jinet_2D():
    input_array = np.zeros((1, 64, 64, 1), dtype=np.float32)
    model2d = JINetModel((64, 64, 1), spacetime_ndim=2)
    result = model2d.predict([input_array])
    assert result.shape == input_array.shape
    assert result.dtype == input_array.dtype


def test_supervised_3D():
    input_array = np.zeros((1, 64, 64, 64, 1), dtype=np.float32)
    model3d = UNetModel(
        (64, 64, 64, 1),
        nb_unet_levels=2,
        shiftconv=False,
        supervised=True,
        spacetime_ndim=3,
    )
    result = model3d.predict(input_array)
    assert result.shape == input_array.shape
    assert result.dtype == input_array.dtype


def test_shiftconv_3D():
    input_array = np.zeros((1, 64, 64, 64, 1), dtype=np.float32)
    model3d = UNetModel(
        (64, 64, 64, 1),
        nb_unet_levels=2,
        shiftconv=True,
        supervised=False,
        spacetime_ndim=3,
    )
    result = model3d.predict(input_array)
    assert result.shape == input_array.shape
    assert result.dtype == input_array.dtype


def test_masking_3D():
    input_array = np.zeros((1, 64, 64, 64, 1), dtype=np.float32)
    model3d = UNetModel(
        (64, 64, 64, 1),
        nb_unet_levels=2,
        shiftconv=False,
        supervised=False,
        spacetime_ndim=3,
    )
    result = model3d.predict([input_array, input_array])
    assert result.shape == input_array.shape
    assert result.dtype == input_array.dtype


def test_various_masking_3D():
    for i in [0, 4]:
        input_array = np.zeros((1, 21 + i, 64, 64, 1), dtype=np.float32)
        print(f'input shape: {input_array.shape}')
        model3d = UNetModel(
            input_array.shape[1:],
            nb_unet_levels=4,
            shiftconv=False,
            supervised=False,
            spacetime_ndim=3,
        )
        result = model3d.predict([input_array, input_array])
        assert result.shape == input_array.shape
        assert result.dtype == input_array.dtype


def test_thin_masking_3D():
    for i in range(3):
        input_array = np.zeros((1, 2 + i, 64, 64, 1), dtype=np.float32)
        print(f'input shape: {input_array.shape}')
        model3d = UNetModel(
            input_array.shape[1:],
            nb_unet_levels=4,
            shiftconv=False,
            supervised=False,
            spacetime_ndim=3,
        )
        result = model3d.predict([input_array, input_array])
        assert result.shape == input_array.shape
        assert result.dtype == input_array.dtype



import numpy

from aydin.nn.util.mask_generator import maskedgen, randmaskgen
from aydin.nn.util.validation_generator import val_data_generator


def get_jinet_fit_args(
    input_image,
    batch_size,
    total_num_patches,
    img_val,
    create_patches_for_validation,
    train_valid_ratio,
):
    """

    Parameters
    ----------
    input_image
    batch_size
    total_num_patches
    img_val
    create_patches_for_validation
    train_valid_ratio

    Returns
    -------
    validation_data

    """
    if create_patches_for_validation:
        val_split = total_num_patches * train_valid_ratio
        val_split = (
            val_split - (val_split % batch_size) + batch_size
        ) / total_num_patches
        validation_data = (
            input_image[int(total_num_patches * val_split) :],
            input_image[int(total_num_patches * val_split) :],
        )
    else:
        validation_data = (input_image, img_val)

    return validation_data


def get_unet_fit_args(
    train_method="random",
    create_patches_for_validation=False,
    input_image=None,
    total_num_patches=None,
    train_valid_ratio=None,
    batch_size=None,
    random_mask_ratio=None,
    img_val=None,
    patch_size=None,
    mask_size=None,
    val_marker=None,
    replace_by=None,
):
    """

    Parameters
    ----------
    train_method
    create_patches_for_validation
    input_image
    total_num_patches
    train_valid_ratio
    batch_size
    random_mask_ratio
    img_val
    patch_size
    mask_size
    val_marker
    replace_by

    Returns
    -------
    validation_data, validation_steps tuple

    """
    validation_steps = None
    validation_data = None

    if 'shiftconv' in train_method:
        if create_patches_for_validation:
            val_split = total_num_patches * train_valid_ratio
            val_split = (
                val_split - (val_split % batch_size) + batch_size
            ) / total_num_patches
            validation_data = (
                input_image[int(total_num_patches * val_split) :],
                input_image[int(total_num_patches * val_split) :],
            )
        else:
            validation_data = (input_image, img_val)
    elif 'checkerbox' in train_method:
        validation_steps = max(
            numpy.floor(input_image.shape[0] * train_valid_ratio / batch_size).astype(
                int
            ),
            1,
        )
        if create_patches_for_validation:
            validation_data = maskedgen(
                input_image, batch_size, mask_size, replace_by=replace_by
            )
        else:
            validation_data = val_data_generator(
                input_image,
                img_val,
                val_marker,
                batch_size,
                train_valid_ratio=train_valid_ratio,
            )
    elif 'random' in train_method:
        validation_steps = max(
            numpy.floor(total_num_patches * train_valid_ratio / batch_size).astype(int),
            1,
        )
        if create_patches_for_validation:
            validation_data = randmaskgen(
                input_image,
                batch_size,
                p_maskedpixels=random_mask_ratio,
                replace_by=replace_by,
            )
        else:
            validation_data = val_data_generator(
                input_image,
                img_val,
                val_marker,
                batch_size,
                train_valid_ratio=train_valid_ratio,
            )

    return validation_data, validation_steps



# flake8: noqa
from random import randint

import napari
import numpy as np
from skimage.draw import line
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import add_noise
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor


def generate_lines_image(
    image_size=1000, number_of_lines=1200, min_line_length=15, max_line_length=50
):
    img = np.zeros((image_size, image_size), dtype=np.float32)

    # draw line
    def add_line(img, r0, c0, r1, c1):
        rr, cc = line(r0, c0, r1, c1)
        img[rr, cc] = 1.0

    # add horizontal lines on top half
    for _ in range(number_of_lines // 2):
        line_length = randint(min_line_length, max_line_length)
        col = randint(0, 949 - line_length)
        row = randint(0, 500)
        add_line(img, row, col, row, col + line_length)

    # add vertical lines on bottom half
    for _ in range(number_of_lines // 2):
        line_length = randint(min_line_length, max_line_length)
        col = randint(0, 999)
        row = randint(500, 949 - line_length)
        add_line(img, row, col, row + line_length, col)

    return img


def demo(image):
    # imsave("ss5.tif", img)

    noisy = add_noise(image, intensity=5, variance=0.1, sap=0.2)

    # imsave("ss5_noisy.tif", img)

    # run wsf:
    generator = StandardFeatureGenerator(
        max_level=8,
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )
    regressor = CBRegressor(patience=30)
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.train(noisy, noisy)
    denoised_wsf = it.translate(noisy)

    # run wsf on flipped image
    flipped_img = np.flipud(noisy)
    denoised_wsf_flipped = np.flipud(it.translate(flipped_img))

    # run:
    generator = StandardFeatureGenerator()
    regressor = CBRegressor(patience=30)
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.train(noisy, noisy)
    denoised = it.translate(noisy)

    # stats
    image_clipped = np.clip(image, 0, 1)
    noisy_clipped = np.clip(noisy, 0, 1)
    psnr_noisy = psnr(image_clipped, noisy_clipped)
    ssim_noisy = ssim(image_clipped, noisy_clipped)

    denoised_wsf_clipped = np.clip(denoised_wsf, 0, 1)
    psnr_denoised_wsf_clipped = psnr(image_clipped, denoised_wsf_clipped)
    ssim_denoised_wsf_clipped = ssim(image_clipped, denoised_wsf_clipped)

    denoised_clipped = np.clip(denoised, 0, 1)
    psnr_denoised_clipped = psnr(image_clipped, denoised_clipped)
    ssim_denoised_clipped = ssim(image_clipped, denoised_clipped)

    denoised_wsf_flipped_clipped = np.clip(denoised_wsf_flipped, 0, 1)
    psnr_denoised_wsf_flipped_clipped = psnr(
        image_clipped, denoised_wsf_flipped_clipped
    )
    ssim_denoised_wsf_flipped_clipped = ssim(
        image_clipped, denoised_wsf_flipped_clipped
    )

    print("noisy   :", psnr_noisy, ssim_noisy)
    print("denoised_wsf:", psnr_denoised_wsf_clipped, ssim_denoised_wsf_clipped)
    print("denoised:", psnr_denoised_clipped, ssim_denoised_clipped)
    print(
        "denoised_wsf_flipped:",
        psnr_denoised_wsf_flipped_clipped,
        ssim_denoised_wsf_flipped_clipped,
    )

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(denoised_wsf, name='denoised_wsf')
        viewer.add_image(denoised, name='denoised')
        viewer.add_image(denoised_wsf_flipped, name='denoised_wsf_flipped')

        # imsave("ss5_wsf.tif", denoised_wsf)
        # imsave("ss5.tif", denoised)


if __name__ == '__main__':
    np.random.seed(0)
    image = generate_lines_image()
    demo(image)



# flake8: noqa
import time

import numpy
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.image_metrics import mutual_information, spectral_mutual_information
from aydin.io.datasets import normalise, add_noise, characters, add_blur_2d, newyork
from aydin.it.deconvolution.lr_deconv_scipy import ImageTranslatorLRDeconvScipy
from aydin.util.log.log import Log


def demo(image):
    Log.enable_output = True

    image = normalise(image.astype(numpy.float32))
    blurred_image, psf_kernel = add_blur_2d(image)
    noisy_and_blurred_image = add_noise(
        blurred_image, intensity=10000, variance=0.0001, sap=0.0000001
    )

    lr = ImageTranslatorLRDeconvScipy(psf_kernel=psf_kernel, max_num_iterations=200)

    start = time.time()
    lr.train(noisy_and_blurred_image, noisy_and_blurred_image)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    lr_deconvolved_image = lr.translate(noisy_and_blurred_image)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    print("Below in order: PSNR, norm spectral mutual info, norm mutual info, SSIM: ")

    def printscore(header, val1, val2, val3, val4):
        print(f"{header}: \t {val1:.4f} \t {val2:.4f} \t {val3:.4f} \t {val4:.4f}")

    printscore(
        "n&b image",
        psnr(image, noisy_and_blurred_image),
        spectral_mutual_information(image, noisy_and_blurred_image),
        mutual_information(image, noisy_and_blurred_image),
        ssim(image, noisy_and_blurred_image),
    )
    # printscore(
    #     "den image",
    #     psnr(image, denoised_image),
    #     spectral_mutual_information(image, denoised_image),
    #     mutual_information(image, denoised_image),
    #     ssim(image, denoised_image),
    # )
    printscore(
        "lr      ",
        psnr(image, lr_deconvolved_image),
        spectral_mutual_information(image, lr_deconvolved_image),
        mutual_information(image, lr_deconvolved_image),
        ssim(image, lr_deconvolved_image),
    )

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(blurred_image, name='blurred')
        viewer.add_image(noisy_and_blurred_image, name='noisy')
        viewer.add_image(lr_deconvolved_image, name='lr_deconvolved_image')


demo(newyork())
demo(characters())



# flake8: noqa
import time

import numpy as np
from skimage.exposure import rescale_intensity

from aydin.io import io
from aydin.io.datasets import examples_single, add_noise
from aydin.it.cnn import ImageTranslatorCNN

image_path = examples_single.royerlab_hcr.get_path()
image, metadata = io.imread(image_path)  # (13, 520, 696)
print(image.shape)
# image = image[0:10, 15:35, 130:167, 130:177]
image = image[:, :, :, :, 100:300, 100:600, 200:700, :]
image = rescale_intensity(image.astype(np.float32), in_range='image', out_range=(0, 1))
noisy = add_noise(image)
max_epochs = 10

it = ImageTranslatorCNN(
    training_architecture='random',
    model_architecture='jinet',
    nb_unet_levels=3,
    batch_norm='instance',  # None,  #
    activation='ReLU',
    mask_size=3,
    max_epochs=max_epochs,
)

it.verbose = 1

start = time.time()
it.train(noisy, image, batch_axes=metadata.batch_axes)
stop = time.time()
print(f"Training: elapsed time:  {stop - start} ")

image = image[:, :, :, 1:2, :, :, :, :]

start = time.time()
denoised = it.translate(
    image,
    batch_axes=metadata.batch_axes,
    tile_size=128,  # image.shape[1:-1],  # [12, 12, 12],  # min(image.shape[1:-1])
)
stop = time.time()
print(f"inference: elapsed time:  {stop - start} ")
import napari

with napari.gui_qt():
    viewer = napari.Viewer()
    viewer.add_image(image, name='image')
    viewer.add_image(
        rescale_intensity(denoised, in_range='image', out_range=(0, 1)), name='denoised'
    )

# demo(image)



import random

import numpy
import scipy


def correlation_distance(input_image, target_image=None, method: str = 'firstmin'):
    """Computes correlation distances

    Parameters
    ----------
    input_image : numpy.typing.ArrayLike
    target_image : numpy.typing.ArrayLike
    method : str

    Returns
    -------
    Tuple of correlation_distances : tuple

    """

    correlation_curves_list = correlation(input_image, target_image)

    correlation_distances_list = []

    for correlation_curve in correlation_curves_list:

        if correlation_curve is None:
            correlation_distances_list.append(0)
        else:
            length = correlation_curve.shape[0]

            if method == 'zerocross':
                correlation_distance = 0
                for distance in range(length):
                    value = correlation_curve[distance]
                    if value < 0:
                        correlation_distance = distance
                        break

                correlation_distances_list.append(correlation_distance)

            elif method == 'firstmin':
                last_value = 1
                correlation_distance = 0
                for distance in range(length):
                    value = correlation_curve[distance]
                    if value > last_value and value < 0:
                        correlation_distance = distance - 1
                        break
                    last_value = value

                correlation_distances_list.append(correlation_distance)

            elif method == 'min':
                min_value = 1
                min_distance = 0
                for distance in range(length):
                    value = correlation_curve[distance]
                    if value < min_value and value < 0:
                        min_value = value
                        min_distance = distance
                correlation_distances_list.append(min_distance)

    return tuple(correlation_distances_list)


def correlation(
    input_image,
    target_image=None,
    nb_samples: int = 4 * 1024,
    max_length: int = 256,
    smooth: bool = True,
):
    """Computes correlation

    Parameters
    ----------
    input_image : numpy.typing.ArrayLike
    target_image : numpy.typing.ArrayLike
    nb_samples : int
    max_length : int
    smooth : bool

    Returns
    -------
    Tuple of correlations : tuple

    """

    # Determine image(s)  shape:
    shape = input_image.shape

    # Initialise target image if  None:
    if target_image is None:
        target_image = input_image

    # Makes sure that the images have the same shape:
    if input_image is not target_image and input_image.shape != target_image.shape:
        raise ValueError('Input image and target image has different shapes.')

    # Number of dimensions:
    nb_dim = len(shape)

    # This list will contain the correlation vectors for each and every dimension:
    corr_list = []

    # We iterate for each dimension:
    for dim in range(nb_dim):

        dim_length = shape[dim]

        if dim_length >= 3:

            max_length_dim = min(dim_length, max_length)

            corr_samples_list = []
            counter = 0

            for sample in range(nb_samples):
                slice_list = list(random.randrange(0, shape[i]) for i in range(nb_dim))

                pos = random.randrange(0, 1 + shape[dim] - max_length_dim)
                slice_list[dim] = slice(pos, pos + max_length_dim, 1)

                line_array_input = input_image[tuple(slice_list)]
                line_array_target = target_image[tuple(slice_list)]

                line_array_input = line_array_input.astype(numpy.float, copy=False)
                line_array_target = line_array_target.astype(numpy.float, copy=False)

                line_array_input = line_array_input - (
                    line_array_input.sum() / line_array_input.shape[0]
                )
                line_array_target = line_array_target - (
                    line_array_target.sum() / line_array_target.shape[0]
                )

                corr = numpy.correlate(line_array_input, line_array_target, mode='full')
                corr = corr[corr.size // 2 :]

                if corr[0] <= 0:
                    continue

                # corr = numpy.abs(corr)

                corr_samples_list.append(corr)

                counter += 1

            if len(corr_samples_list) > 0:
                corr_samples_stack = numpy.stack(corr_samples_list)
                corr_avg = numpy.median(corr_samples_stack, axis=0)

                # corr_avg = corr_avg / numpy.sum(corr_avg)

                if smooth and corr_avg.size >= 3:
                    corr_avg[1:] = numpy.convolve(
                        corr_avg, numpy.ones(3) / 3.0, mode='same'
                    )[1:]
                    corr_avg[1:] = numpy.convolve(
                        corr_avg, numpy.ones(3) / 3.0, mode='same'
                    )[1:]
                    corr_avg[1:] = scipy.signal.medfilt(corr_avg, kernel_size=3)[1:]
                    corr_avg[1:] = scipy.signal.medfilt(corr_avg, kernel_size=5)[1:]
                    corr_avg[1:] = scipy.signal.medfilt(corr_avg, kernel_size=7)[1:]
                    # corr_avg[1:] = numpy.convolve(corr_avg, numpy.ones(3) / 3.0, mode='same')[1:]
                    # corr_avg[1:] = numpy.convolve(corr_avg, numpy.ones(3) / 3.0, mode='same')[1:]
                    # corr_avg = numpy.convolve(corr_avg, numpy.ones(3) / 3.0, mode='same')

                corr_avg = corr_avg / corr_avg[0]
            else:
                corr_avg = None

        else:
            # Dimension is way too short:
            corr_avg = None

        corr_list.append(corr_avg)

    return tuple(corr_list)



from aydin.analysis.correlation import correlation_distance
from aydin.io import io
from aydin.io.datasets import examples_single


def demo_analysis():
    for example in examples_single:
        example_file_path = example.get_path()

        # print(f"Trying to open and make sense of file {example_file_path}")

        array, metadata = io.imread(example_file_path)
        print(f"File        :  {example}")
        print(f"Metadata    :  {metadata}")
        print(f"Array shape :  {array.shape}")
        print(f"Array dtype :  {array.dtype}")

        correlations = correlation_distance(array)
        print(f"Correlations:  {correlations} ")


demo_analysis()



from functools import reduce
from math import sqrt
from operator import mul
from typing import Tuple, Union, Optional
import matplotlib.pyplot as plt
import numpy as np
from sklearn.cluster import MiniBatchKMeans
from sklearn.feature_extraction.image import _extract_patches

from aydin.util.log.log import lprint, lsection


def extract_kernels(
    image,
    size: int = 7,
    num_kernels: int = None,
    num_patches: int = 1e5,
    num_iterations: int = 100,
    display: bool = False,
):
    if num_kernels is None:
        num_kernels = size ** image.ndim

    # #############################################################################
    # Learn the dictionary of images

    with lsection(
        f'Learning a dictionary of {num_kernels} kernels of size {size} from {num_patches} patches'
    ):
        rng = np.random.RandomState(0)
        kmeans = MiniBatchKMeans(n_clusters=num_kernels, random_state=rng, verbose=True)
        kernel_size = (size,) * image.ndim

        # The online learning part: cycle over the whole dataset 6 times
        for i in range(num_iterations):
            lprint(f"Iteration: {i}")
            data = extract_patches_nd(
                image, kernel_size, num_patches=int(num_patches) // num_iterations
            )
            data = np.reshape(data, (len(data), -1))
            data -= np.min(data, axis=0)
            data /= np.sum(data, axis=0)
            kmeans.partial_fit(data)

    if display:
        # #############################################################################
        # Plot the results
        plt.figure(figsize=(8, 8))
        for i, kernel in enumerate(kmeans.cluster_centers_):
            s = sqrt(num_kernels)
            plt.subplot(s + 1, s + 1, i + 1)
            plt.imshow(
                kernel.reshape(kernel_size),
                cmap=plt.cm.gray,
                interpolation='nearest',  # , vmin=0, vmax=1
            )
            plt.xticks(())
            plt.yticks(())

        plt.subplots_adjust(0.08, 0.02, 0.92, 0.85, 0.08, 0.23)

        plt.show()

    return [np.reshape(kernel, kernel_size) for kernel in kmeans.cluster_centers_]


def extract_patches_nd(
    image, patch_shape: Union[int, Tuple[int, ...]], num_patches: Optional[int] = None
):
    """Reshape a 2D image into a collection of patches

    The resulting patches are allocated in a dedicated array.

    Read more in the :ref:`User Guide <image_feature_extraction>`.

    Parameters
    ----------
    image : ndarray of shape (image_height, image_width) or \
        (image_height, image_width, n_channels)
        The original image data. For color images, the last dimension specifies
        the channel: a RGB image would have `n_channels=3`.

    patch_shape : int or tuple of ints (..., depth, heigt, width)
        The shape of one patch. Can be a single int for all axis.

    num_patches : int or float, default=None
        The maximum number of patches to extract. If `max_patches` is a float
        between 0 and 1, it is taken to be a proportion of the total number
        of patches.

    Returns
    -------
    patches : array of shape (n_patches, patch_height, patch_width) or \
        (n_patches, patch_height, patch_width, n_channels)
        The collection of patches extracted from the image, where `n_patches`
        is either `max_patches` or the total number of patches that can be
        extracted.
    """

    if type(patch_shape) is not tuple:
        patch_shape = (patch_shape,) * image.ndim

    extracted_patches = _extract_patches(
        image, patch_shape=patch_shape, extraction_step=1
    )

    extracted_patches_shape = extracted_patches.shape[: image.ndim]
    num_extracted_patches = prod(extracted_patches_shape)

    if num_patches is not None:
        if type(num_patches) is float:
            num_patches = int(num_patches * num_extracted_patches)

        if type(num_patches) is int:
            num_patches = min(num_patches, num_extracted_patches)

        p = num_patches / num_extracted_patches
        indices = np.random.choice(
            a=[False, True], size=extracted_patches_shape, p=[1 - p, p]
        )
        patches = extracted_patches[indices]

    else:
        patches = extracted_patches

    patches = patches.reshape(-1, *patch_shape)

    return patches


def prod(atuple: Tuple[Union[float, int]]):
    # In python 3.8 there is a prod function in math, until then we have:
    return reduce(mul, atuple)



# flake8: noqa
import numpy
import numpy as np

from aydin.io.datasets import normalise, examples_single
from aydin.it.classic_denoisers.butterworth import calibrate_denoise_butterworth
from aydin.util.log.log import Log, lprint


def demo_lowpass_islet(noisy, display=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    Log.set_log_max_depth(5)

    lprint(noisy.shape)
    noisy = normalise(noisy.astype(np.float32))

    function, parameters, _ = calibrate_denoise_butterworth(
        noisy, isotropic=False, axes=(0, 1, 2)
    )
    denoised = function(noisy, **parameters)

    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(noisy, name='noisy')
            viewer.add_image(denoised, name='denoised')


if __name__ == "__main__":
    islet = examples_single.royerlab_hcr.get_array().squeeze()
    # islet = islet[2, :, 400: 400 + 256, 700: 700 + 256]
    islet = islet[1, :500, :500]
    demo_lowpass_islet(islet)

    # hyman_hela = examples_single.hyman_hela.get_array()
    # demo_lowpass_anisotropic(hyman_hela)



import pytest
import numpy as np
from tifffile import imread

from aydin.util.misc.resource import download_and_extract_zipresource


@pytest.mark.heavy
def test_download_and_extract_zipresource():

    # Testing execution of the method
    try:
        download_and_extract_zipresource(
            url='http://csbdeep.bioimagecomputing.com/example_data/tribolium.zip',
            targetdir='data',
        )
    except Exception:
        pytest.fail("download_and_extract_zipresource failed", True)

    # Read some files
    x = imread('data/tribolium/train/GT/nGFP_0.1_0.2_0.5_20_13_late.tif').astype(
        np.float32, copy=False
    )
    y = imread('data/tribolium/train/low/nGFP_0.1_0.2_0.5_20_13_late.tif').astype(
        np.float32, copy=False
    )
    z = imread('data/tribolium/test/GT/nGFP_0.1_0.2_0.5_20_14_late.tif').astype(
        np.float32, copy=False
    )
    t = imread('data/tribolium/test/low/nGFP_0.1_0.2_0.5_20_14_late.tif').astype(
        np.float32, copy=False
    )

    # Check if dimensions of images match from different files, we know they should be same
    assert y.shape == x.shape
    assert z.shape == t.shape



from aydin.io.datasets import cropped_newyork

from aydin.it.classic_denoisers.demo.demo_2D_butterworth import demo_butterworth
from aydin.it.classic_denoisers.butterworth import denoise_butterworth
from aydin.it.classic_denoisers.test.util_test_nd import check_nd


def test_butterworth():
    assert demo_butterworth(cropped_newyork(), display=False) >= 0.608 - 0.03


def test_butterworth_nd():
    check_nd(denoise_butterworth)



import numpy
import scipy
from scipy.fft import next_fast_len
from scipy.signal.signaltools import _init_freq_conv_axes, _apply_conv_mode

from aydin.it.deconvolution.lr_deconv import ImageTranslatorLRDeconv


class ImageTranslatorLRDeconvScipy(ImageTranslatorLRDeconv):
    """Lucy Richardson Deconvolution, SciPy backend"""

    def __init__(self, *args, **kwargs):
        """Constructs a Lucy Richardson deconvolution image translator.

        Parameters
        ----------
        psf_kernel : numpy.typing.ArrayLike
            2D or 3D kernel, dimensions should be odd numbers and numbers sum to 1
        max_num_iterations : int
        clip : bool
        backend : str
            Computation backend selection.
        kwargs : dict
        """
        super().__init__(*args, **kwargs)

        self.max_voxels_per_tile = (
            512 ** 3 if self.max_voxels_per_tile is None else self.max_voxels_per_tile
        )

    def _convert_array_format_in(self, input_image):
        return input_image

    def _convert_array_format_out(self, output_image):
        return output_image

    def _get_convolution_method(self, input_image, psf_kernel):
        """
        Method to get convolution method.

        Parameters
        ----------
        input_image : array_like
        psf_kernel : array_like

        Returns
        -------
        convolve

        """

        return _fftconvolve

    def _get_pad_method(self):
        """
        Method to get pad method.

        Returns
        -------
        pad

        """
        import numpy

        return numpy.pad


def _fftconvolve(in1, in2, mode="full", axes=None):
    in1 = numpy.asarray(in1)
    in2 = numpy.asarray(in2)

    if in1.ndim == in2.ndim == 0:  # scalar inputs
        return in1 * in2
    elif in1.ndim != in2.ndim:
        raise ValueError("in1 and in2 should have the same dimensionality")
    elif in1.size == 0 or in2.size == 0:  # empty arrays
        return numpy.array([])

    in1, in2, axes = _init_freq_conv_axes(in1, in2, mode, axes, sorted_axes=False)

    s1 = in1.shape
    s2 = in2.shape

    shape = [
        max((s1[i], s2[i])) if i not in axes else s1[i] + s2[i] - 1
        for i in range(in1.ndim)
    ]

    ret = _freq_domain_conv(in1, in2, axes, shape, calc_fast_len=True)

    return _apply_conv_mode(ret, s1, s2, mode, axes)


def _freq_domain_conv(in1, in2, axes, shape, calc_fast_len=False):
    if not len(axes):
        return in1 * in2

    complex_result = in1.dtype.kind == 'c' or in2.dtype.kind == 'c'

    if calc_fast_len:
        # Speed up FFT by padding to optimal size.
        fshape = [next_fast_len(shape[a], not complex_result) for a in axes]
    else:
        fshape = shape

    if not complex_result:
        fft, ifft = scipy.fft.rfftn, scipy.fft.irfftn
    else:
        fft, ifft = scipy.fft.fftn, scipy.fft.ifftn

    sp1 = fft(in1, fshape, axes=axes, workers=-1)
    sp2 = fft(in2, fshape, axes=axes, workers=-1)

    ret = ifft(sp1 * sp2, fshape, axes=axes, workers=-1)

    if calc_fast_len:
        fslice = tuple([slice(sz) for sz in shape])
        ret = ret[fslice]

    return ret



import numpy
from numba import jit
from scipy.ndimage import median_filter, uniform_filter

__fastmath = {'contract', 'afn', 'reassoc'}
__error_model = 'numpy'


def calibrate_denoise_lipschitz(
    image,
    lipschitz: float = 0.1,
    percentile: float = 0.001,
    alpha: float = 0.1,
    max_num_iterations: int = 128,
    **other_fixed_parameters,
):
    """
    Calibrates the Lipschitz denoiser for the given image and returns the
    optimal parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate Sobolev denoiser for.

    lipschitz : float
        Increase to tolerate more variation, decrease to be
        more aggressive in removing salt & pepper noise.

    percentile : float
        Percentile value used to determine the threshold
        for choosing the worst offending voxels per iteration.
        (advanced)

    alpha : float
        This constant controls the amount of correction per iteration.
        Should be a number within[0, 1]. (advanced)

    max_num_iterations: int
        Maximum number of Lipschitz correction iterations to run. (advanced)


    other_fixed_parameters: dict
        Any other fixed parameters. (advanced)

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # Combine fixed parameters:
    other_fixed_parameters = other_fixed_parameters | {
        'lipschitz': lipschitz,
        'percentile': percentile,
        'alpha': alpha,
        'max_num_iterations': max_num_iterations,
    }

    best_parameters = other_fixed_parameters

    # Memory needed:
    memory_needed = 3 * image.nbytes  # complex numbers and more

    return denoise_lipschitz, best_parameters, memory_needed


def denoise_lipschitz(
    image,
    lipschitz: float = 0.05,
    percentile: float = 0.01,
    alpha: float = 0.1,
    max_num_iterations: int = 128,
    multi_core: bool = True,
):
    """
    Denoises the given image by correcting voxel values that violate
    <a href="https://en.wikipedia.org/wiki/Lipschitz_continuity">Lipschitz
    continuity</a> criterion. These voxels get replaced
    iteratively replaced by the median. This is repeated for a number
    of iterations (max_num_iterations), until no more changes to
    the image occurs, and making sure that the proportion of
    corrected voxels remains below a given level (max_corrections argument).


    Parameters
    ----------
    image: ArrayLike
        Image to be denoised

    lipschitz : float
        Increase to tolerate more variation, decrease to be
        more aggressive in removing impulse/salt&pepper noise.

    percentile : float
        Percentile value used to determine the threshold
        for choosing the worst offending voxels per iteration.
        (advanced)

    alpha : float
        This constant controls the amount of correction per ietartion.
        Should be a number within[0, 1]. (advanced)

    max_num_iterations: int
        Maximum number of Lipschitz correction iterations to run. (advanced)

    multi_core: bool
        By default we use as many cores as possible, in some cases, for small
        (test) images, it might be faster to run on a single core instead of
        starting the whole parallelization machinery. (advanced)


    Returns
    -------
    Denoised image

    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=True)

    # Compute median:
    median_ref = median_filter(image, size=3)

    # Wrap compute_error function:
    wrapped_compute_error = jit(nopython=True, parallel=multi_core)(_compute_error)

    for i in range(max_num_iterations):
        # lprint(f"Iteration {i}")

        # Compute median:
        median = uniform_filter(image, size=5)

        # We scale the lipschitz threshold to the image std at '3 sigma' :
        _lipschitz = lipschitz * 6 * image.std()

        # We compute the 'error':
        error = wrapped_compute_error(image, median=median, lipschitz=_lipschitz)

        # We compute the threshold on the basis of the errors,
        # we first tackle the most offending voxels:
        threshold = numpy.percentile(error, q=100 * (1 - percentile))

        # We compute the mask:
        mask = error > threshold

        # count number of corrections for this round:
        num_corrections = numpy.sum(mask)
        # lprint(f"Number of corrections: {num_corrections}")

        # if no corrections made we stop iterating:
        if num_corrections == 0:
            break

        # We use the median to correct pixels:
        image[mask] = (alpha) * image[mask] + (1 - alpha) * median_ref[mask]

    return image


def _compute_error(array, median, lipschitz):
    # we compute the error map:
    error = median - array
    error = numpy.abs(error)
    error = numpy.maximum(error, lipschitz)
    error -= lipschitz
    return error



# flake8: noqa
import numpy
import numpy as np
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import (
    normalise,
    add_noise,
    newyork,
    camera,
    characters,
    pollen,
    lizard,
    dots,
)
from aydin.it.classic_denoisers.tv import calibrate_denoise_tv
from aydin.util.log.log import Log


def demo_tv(image, display=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    Log.set_log_max_depth(5)

    image = normalise(image.astype(np.float32))
    noisy = add_noise(image)

    function, parameters, memreq = calibrate_denoise_tv(noisy)
    denoised = function(noisy, **parameters)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)
    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)
    print("   noisy   :", psnr_noisy, ssim_noisy)
    print("tv denoised:", psnr_denoised, ssim_denoised)

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(noisy, name='noisy')
            viewer.add_image(denoised, name='denoised')

    return ssim_denoised


if __name__ == "__main__":
    newyork_image = newyork()
    demo_tv(newyork_image)
    characters_image = characters()
    demo_tv(characters_image)
    pollen_image = pollen()
    demo_tv(pollen_image)
    lizard_image = lizard()
    demo_tv(lizard_image)
    dots_image = dots()
    demo_tv(dots_image)
    camera_image = camera()
    demo_tv(camera_image)



def human_readable_byte_size(nbytes: int) -> str:
    """

    Parameters
    ----------
    nbytes : int

    Returns
    -------
    Human readable string : str

    """
    suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
    i = 0
    while nbytes >= 1024 and i < len(suffixes) - 1:
        nbytes /= 1024.0
        i += 1
    f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
    return f"{f} {suffixes[i]}"
